unit UGoal;

// A "goal" compares 2 states and calc a "score" to be added to the future.

interface

uses UState, ExtCtrls;


type TGoal = class(TObject)
  private
    // Owner of the gaol: Can be a player or a team.
    fOwner: TObject;
    // First, to show it to the user I need a name:
    fDisplayName: string;
    //
    // Defining a goal consist in giving value to those params:
    //
    // First, you have some system quantity that the feeling is about, it can be
    // your speed if you love speeding, your health level or anything you can
    // compute inside your agent simulation:
    fQuantity: double;
    // Gain / Loose feelings need to use the previous value of the Quantity:
    fOldQuantity: double;
    // Minimum > 0 makes coefs never be zero (energy):
    fMinimum: double;
    // Whener Quantity changes, the feeling outputs must be recalculated:
    FeelingNeedRecalc: boolean;
    // How "strong" the goal feelings are overall:
    fStrength, fMaxStrength: double;
    // Asymetric feelings:
    fAsymmetry: double;
    // Goals with a Gain need to define a reaction time: For instance
    // health decreasing fires a Gain < 1 so ReactionMSeconds of 5000 means that
    // you are to be warned 5 seconds before a crash to have time to react:
    fReactionMSeconds: integer;
    // Once the FinalCoef fires at ReactionTime, it can decrease faster or
    // slower as times tends to zero.
    fUrgency: double;
    //
    // A goal scores at each time step giving you 3 outputs:
    //
    // Enjoy feeling is the main scoring:
    fEnjoy: double;
    // Mood coef. changes the "enjoy" at each step as the quantity reach a nominal
    // level, like health is limited by 100% so Mood=Health/100, for intance:
    fMoodCoef, fNominalQuantity: double;
    // Looses/Gains coef. that changes the whole future score:
    fGain: double;
    //
    // Scorings of a goal can be shared to others players multiplied by a coef.
    // so several players can cooperate (positive coef) or compete (negative),
    // with a value of 0 meaning that the goal is "personal" (not shared).
    //
    // As other players can be of my team or not, 2 coefs. are defined, one
    // applied to "friends" (in my team) and another for "foes" (all others):
    //
    // Note: Not used in this particular version.
    fShareWithFriends, fShareWithFoes: double;
    //
    function  GetSimSeconds: double;
    procedure SetStrength(v: double);
    function  GetQuantity: double; virtual;
    function  GetOldQuantity: double;
    procedure SetQuantity(v: double); virtual;
    // Reaction time in seconds instead of milli seconds:
    procedure SetReactionTime(v: double);
    function  GetReactionTime: double;
    procedure CalcFeeling(v: TObject); virtual;
  protected
    procedure DrawOnCalc; virtual;
    // THIS is the function to define for each goal:
    procedure ReadQuantity(v: TObject); virtual; abstract;
    property  Quantity: double read GetQuantity write SetQuantity;
    property  OldQuantity: double read GetOldQuantity write fOldQuantity;
    property  Minimum: double read fMinimum write fMinimum;
  public
    constructor create(Sender: TObject); virtual;
    // Apply the goal feelings/scoring to a player's state:
    procedure ApplyToPlayer(v: TObject);
    // Goals with a "state", like energy drops, needs a second "state" used
    // while imagining that need to be reset on starting a new future.
    procedure Reset; virtual;
    // Draw the effect of the goal after a new frame is done:
    procedure DrawOnFrame; virtual;
    // Copy from other goal:
    procedure CopyFrom(g: TGoal);
    //
    property  DisplayName: string read fDisplayName write fDisplayName;
    property  Owner: TObject read fOwner write fOwner;
    // Feeling input params:
    property  Strength: double read fStrength write SetStrength;
    property  MaxStrength: double read fMaxStrength write fMaxStrength;
    property  Asymmetry: double read fAsymmetry write fAsymmetry;
    property  NominalQuantity: double read fNominalQuantity write fNominalQuantity;
    property  ReactionTime: double read GetReactionTime write SetReactionTime;
    property  ReactionMSeconds: integer read fReactionMSeconds write fReactionMSeconds;
    property  Urgency: double read fUrgency write fUrgency;
    // Feeling outputs values:
    property  Enjoy:    double read fEnjoy    write fEnjoy;
    property  MoodCoef: double read fMoodCoef write fMoodCoef;
    property  Gain: double read fGain write fGain;
    // Reading track params:
    property  SimSeconds: double read GetSimSeconds;
  end;


implementation

uses Math, UPlayer, UTeam,  UTrack;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TGoal.create(Sender: TObject);
begin
  inherited create;
  // Owner of the goal (a single Player or a complete Team of players)
  fOwner:= Sender;
  fDisplayName:= 'Generic goal';
  // Global strength:
  MaxStrength:= 5.0;
  Strength:= 1;
  fAsymmetry:= 1; // Negatives are just as strong as positives
  // Mood feeling:
  NominalQuantity:= 0;
  Minimum:= 0;
  // Gain feeling:
  ReactionMSeconds:= 0;
  Urgency:= 0.5;
  // Cooperative:
  fShareWithFriends:= 0;
  fShareWithFoes:= 0;
end;

procedure TGoal.CopyFrom(g: TGoal);
begin
  // General strength:
  Strength:=  g.Strength;
  MaxStrength:= g.MaxStrength;
  // Happy feeling: Nominal quantity for Happy = 1 unit.
  NominalQuantity:= g.NominalQuantity;
  // Gain feelings:
  ReactionMSeconds:= g.ReactionMSeconds;
  Urgency:= g.Urgency;
  // Shared feelings (cooperative hives):
  fShareWithFriends:= g.fShareWithFriends;
  fShareWithFoes:= g.fShareWithFoes;
end;

// Gain/Loose coef. accumulate during a future, so before starting a new
// future it must be reset. Other goals would reset other things too.
procedure TGoal.Reset;
begin
  Gain:= 1;
end;


// *******************
// ** FEELING STUFF **
// *******************

procedure TGoal.ApplyToPlayer(v: TObject);
begin
  // Un animated players don't have goals, but they will if you remove the line!
  if not TPlayer(v).Animated then
    exit;
  // Define old and actual Quantity from the player so the goal can calculate
  // the associated feelings:
  ReadQuantity(v);
  // Calculate scorings for this player comparing old and actual state:
  CalcFeeling(v);
  // Add scorings to the future:
  TPlayer(v).State.AddGoalScoring(self);
  // Draw efect?
  if TPlayer(v).GoalSet.Debug then
    DrawOnCalc;
end;

// Each goal need to define this quantity by its own:
// NOTE: The OldQuantity need to be setted BEFORE the Quantity, so the Gain or
//       Loose associated to the feeling can be computed (use ApplyToPlayer):
function TGoal.GetQuantity: double;
begin
  result:= fQuantity;
end;
// When quantity changes, the output feelings need to be recalculated:
procedure TGoal.SetQuantity(v: double);
begin
  fQuantity:= v;
  FeelingNeedRecalc:= true;
end;
function TGoal.GetOldQuantity: double;
begin
  result:= fOldQuantity;
end;

// Recalculate  the feeling outputs after the Quantity feeled has changed:
// All goals works the same way! Knowing Quantity, Strength, etc. a goal
// can calculate its feelings (but you can override it if neccesary):
procedure TGoal.CalcFeeling(v: TObject);
begin
  // 1) Enjoy feeling = quantity "enjoyed" x time enjoyed (dt):
  Enjoy:= Quantity;
  // 2) The mood is a coef that changes the enjoy of all goals present on the
  //    agent, not only the one that generates the "mood" change, and is applied
  //    step by step:
  if (NominalQuantity > 0) then begin
    // Health is 40% and Nominal is 100%, so mood is 40/100 = 0.4
    MoodCoef:= Max(0, Quantity) / NominalQuantity;
  end else begin
    // There is no "mood change" associated with the goal.
    MoodCoef:= 1;
  end;
  // 3) The Gain/Loose coef is only needed is there is a ReactionTime defined.
  //    The coef will be applied to the future total enjoy, once added the
  //    enjoy at each step taken.
  if (ReactionMSeconds > 0) and (Quantity<>OldQuantity) then begin
    if (OldQuantity = 0) then begin
      // You were already dead, so loosing is not possible anymore!
      Gain:= 1;
    end else begin
      Gain:= ((Quantity+Minimum) / (OldQuantity+Minimum));
      if (Quantity < OldQuantity) then begin
        // Loose feeling:
        if fAsymmetry < 1 then
          Gain:= Gain + (1-Gain) * (1-fAsymmetry);
      end else begin
        // Gain feeling:
        if fAsymmetry > 1 then
          Gain:= Gain / fAsymmetry;
      end;
    end;
  end else begin
    Gain:= 1;
  end;
  Gain:= Gain * 1.0;
end;


// *******************
// ** DRAWING STUFF **
// *******************

// Fake draw procedures, override them on each kind of goal when needed:
procedure TGoal.DrawOnCalc;
begin
end;
procedure TGoal.DrawOnFrame;
begin
end;


// *********************
// ** SETTERS/GETTERS **
// *********************

procedure TGoal.SetStrength(v: double);
begin
  fStrength:= Max(0.0, Min(v, MaxStrength));
end;

// Reaction time in seconds instead of milli seconds:
procedure TGoal.SetReactionTime(v: double);
begin
  ReactionMSeconds:= round(v*1000);
end;
function  TGoal.GetReactionTime: double;
begin
  result:= ReactionMSeconds / 1000;
end;

// Get the simulation delta of time (like 0.1 seconds):
function TGoal.GetSimSeconds: double;
begin
  if Owner is TPlayer then
    result:= TPlayer(Owner).StepSeconds
  else
    result:= TTeam(Owner).StepSeconds;
end;

end.
