unit UGoalSet;

// Players and Teams have a "GoalSet", a list of "Goals" they follow.

interface

uses Contnrs, Graphics,
     UUTils, UGoal;


type TGoalSet = class(TObject)
  private
    // Owner of the goal set can be a player or a team of players:
    fOwner:  TObject;
    // Witch goals motivate this agent:
    fGoals: TObjectList;
    // Do you want to see the goals on screen? (energy dropds, etc):
    fDebug: boolean;
    function GetCount: integer;
    function GetGoal(m: integer): TGoal;
  protected

  public
    constructor create(sender: TObject);
    destructor  destroy; override;
    procedure   AddGoal(v: TGoal);
    procedure   DelGoal(g: integer);
    procedure   CopyFrom(Other: TGoalSet);
    procedure   Clear;
    procedure   Reset;
    procedure   RandomizeGoals;
    procedure   MutateGoals;
    procedure   ColorizeOwner;
    procedure   CalcGoals;
    procedure   DrawOnFrame;
    property    Owner: TObject read fOwner write fOwner;
    property    Debug: boolean read fDebug write fDebug;
    property    Goal[m: integer]: TGoal read GetGoal;
    property    Count: integer read GetCount;
  end;


implementation

uses
  UPlayer, UTeam;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TGoalSet.create(Sender: TObject);
begin
  inherited create;
  fOwner:= Sender;
  fGoals:= TObjectList.create;
  fGoals.OwnsObjects:= true;
  fDebug:= false;
end;

destructor TGoalSet.destroy;
begin
  fGoals.Free;
  inherited;
end;

// Reset goal imaginary internal state (only used on some, like energy drops):
procedure TGoalSet.Reset;
var
  g: integer;
begin
  for g:= 0 to Count-1 do
    Goal[g].Reset;
end;

// Delete all goals:
procedure TGoalSet.Clear;
begin
  fGoals.Clear;
end;

procedure TGoalSet.CopyFrom(Other: TGoalSet);
var
  g, og: integer;
begin
  // For each of my goals:
  for g:= 0 to Count-1 do begin
    // Find a similar one (heatlh <-> health) on the other GoalSet:
    for og:= 0 to Other.Count-1 do begin
      if (Goal[g].ClassName = Other.Goal[og].ClassName) then begin
        // Copy all goal internal params:
        Goal[g].CopyFrom(Other.Goal[og]);
        // Only one goal of each type is supposed to exist!
        break;
      end;
    end;
  end;
end;

// Add a new goal:
procedure TGoalSet.AddGoal(v: TGoal);
begin
  fGoals.Add(v);
end;

// Delete a goal:
procedure TGoalSet.DelGoal(g: integer);
begin
  if (g>=0) and (g<Count) then
    fGoals.Delete(g);
end;


// ********************
// ** FEELINGS STUFF **
// ********************

// After a simulation step, the State is changed, and you need to score it by
// calculating the goals' feelings associated with the change:
procedure TGoalSet.CalcGoals;
var
  g, p: integer;
begin
  for g:= 0 to Count-1 do begin
    if Owner is TPlayer then begin
      // It is a personal goal of a player:
      Goal[g].ApplyToPlayer(TPlayer(Owner));
    end else begin
      // It is a team shared goal:
      for p:= 0 to TTeam(Owner).PlayerCount-1 do
        Goal[g].ApplyToPlayer(TTeam(Owner).Player[p]);
    end;
  end;
end;


// *******************
// ** DRAWING STUFF **
// *******************

// Draw the goals after a step is done and I am drwing the new frame:
procedure TGoalSet.DrawOnFrame;
var
  g: integer;
begin
  if Debug then
    for g:= 0 to Count-1 do
      Goal[g].DrawOnFrame;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

// Randomly decide the goals' strengths:
procedure TGoalSet.RandomizeGoals;
var
  g: integer;
begin
  for g:= 0 to Count-1 do begin
    Goal[g].Strength:= (0.1 + 0.8*random) * Goal[g].MaxStrength;
    // All other params should also be change.
  end;
end;

// Randomly change the actual goals' strengths:
procedure TGoalSet.MutateGoals;
var
  g: integer;
const
  Percent: double = 0.1; // +/-10% of change.
begin
  for g:= 0 to Count-1 do
    Goal[g].Strength:= Goal[g].Strength + Goal[g].MaxStrength * Percent * (2*random - 1);
end;

// Use the first 3 goal strengths to colorize the owner so evolution is visible on screen:
procedure TGoalSet.ColorizeOwner;
var
  i, p: integer;
  R: array [0..2] of integer;
  NewColor: TColor;
begin
  // Use first 3 goals strength to define R, G and B components:
  for i:= 0 to 2 do begin
    if (Count > i) then
      R[i]:= round(255 * Goal[i].Strength/Goal[i].MaxStrength)
    else
      R[i]:= 128;
  end;
  NewColor:= RGB2Color(R[0], R[1], R[2]);
  // Apply the color to the owner:
  if Owner is TPlayer then begin
    TPlayer(Owner).Color:= NewColor;
  end else begin
    for p:= 0 to TTeam(Owner).PlayerCount-1 do
      TTeam(Owner).Player[p].Color:= NewColor;
  end;
end;


// *********************
// ** SETTERS/GETTERS **
// *********************

function TGoalSet.GetGoal(m: integer): TGoal;
begin
  result:= TGoal(fGoals[m]);
end;

function TGoalSet.GetCount: integer;
begin
  result:= fGoals.count;
end;


end.
