unit UListOfFutures;

// When a player "imagines" a new possible future, it finally get a "State" that
// contains both the final state of the player (on its computed StrateParam's
// Values) plus the initial decision taken in order to later by able to group
// futures by "options" using.
//
// The main reason to have this unit is to keep all the heart of the intelligence
// out of the other stuff:
//
// 1) Group futures into its corresponding "option".
// 2) Discard futures with same ending point (MacroState) on each option.
// 3) Score options by adding the score of its futures.
// 4) Take the intelligent decision based on options and their scorings.

interface

uses Contnrs, Graphics,
     UState;


type TListOfFutures = class(TObject)
  private
    fOwner: TObject; // is a TPlayer
    // The list of futures states:
    fFutures: TObjectList;
    // List of "options" = different initial decisions in the futures
    fOptions: Array of TObject;
    function GetFuture(i: integer): TState;
    function GetOption(i: integer): TState;
    function GetOptionCount: integer;
    function GetFutureCount: integer;
  protected
  public
    constructor create(v: TObject);
    destructor destroy; override;
    // Add or take futures:
    procedure AddFuture(v: TState);
    procedure Clear;
    // Debugging:
    procedure DrawOptions;
    // Thinking proccess as a whole:
    procedure TakeIntelligentDecision(DestinationState: TState);
    // Now the parts of the proccess:
    procedure ScoreOptions;
    procedure ReduceOptions;
    procedure NormalizeOptions;
    procedure AverageOptions(DestinationState: TState);
    property  Owner: TObject read FOwner;
    // Property to count futures/options:
    property  OptionCount: integer read GetOptionCount;
    property  FutureCount: integer read GetFutureCount;
    // Properties to access a given future/option as a TState:
    property  Future[i: integer]: TState read GetFuture;
    property  Option[i: integer]: TState read GetOption;
  end;


implementation

uses Classes, Math, Forms, SysUtils,
     UPlayer, UTrack;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TListOfFutures.create(v: TObject);
begin
  inherited create;
  fOwner:= v;
  fFutures:= TObjectList.create;
  fFutures.OwnsObjects:= false;
  SetLength(fOptions, 0);
end;

destructor TListOfFutures.destroy;
var
  i: integer;
begin
  // Futures:
  while fFutures.Count > 0 do begin
    TState(fFutures.Items[0]).Free;
    fFutures.Delete(0);
  end;
  fFutures.free;
  // Options:
  for i:= 0 to High(fOptions) do
    TState(fOptions[i]).Free;
  inherited;
end;

// Before starting a new AI cycle, reset the options:
procedure TListOfFutures.Clear;
var
  i: integer;
begin
  // Clean future list:
  while fFutures.Count > 0 do begin
    TState(fFutures.Items[0]).Free;
    fFutures.Delete(0);
  end;
  // Rest score of existing options:
  for i:= Low(fOptions) to High(fOptions) do
    Option[i].ResetScore;
  // Adjust number of possible options (first time only):
  while (High(fOptions) + 1) < (2*TPlayer(Owner).State.FreeParamCount) do begin
    // Add room for a new option:
    SetLength(fOptions, High(fOptions)+2);
    // Populate it with a state:
    fOptions[High(fOptions)]:= TState.create(Owner);
  end;
end;


// ********************
// ** THINKING STUFF **
// ********************

// Building the list of different futures for each option:

// A new future was traced, add it to the list if it is a different one:
procedure TListOfFutures.AddFuture(v: TState);
var
  f: integer;
  NewFuture: TState;
begin
  // Compare with the other futures from the same option:
  for f:= 0 to FutureCount-1 do begin
    if v.HaveSameOption(Future[f]) and v.HaveSameFinalState(Future[f]) then begin
      // I already have this ending point stored, discard duplicated:
      // Before discarting, keep the best scoring for this end point:
      if (v.Score > Future[f].Score) then
        Future[f].CopyScoreFrom(v);
      exit;
    end;
  end;
  // This future is new for me, store it:
  NewFuture:= TState.create(v.owner);
  NewFuture.CopyFrom(v);
  fFutures.Add(NewFuture);
end;

// I have a list of different futures with it scores, I use them to score my
// options and then calculate the AI decision as a weightened average of options.


// Decision taking proccess:
procedure TListOfFutures.TakeIntelligentDecision(DestinationState: TState);
begin
  //
  // Score each option with the different futures belonging to it:
  ScoreOptions;
  // Reduce options so minimum score is zero:
  ReduceOptions;
  // Normalize the option scores so they sum 1:
  NormalizeOptions;
  // Average option*score will produce the standard "intelligent decisions":
  AverageOptions(DestinationState);
end;

procedure TListOfFutures.ScoreOptions;
var
  f, o: integer;
begin
  for f:= 0 to FutureCount-1 do begin
    o:= Future[f].OptionId;
    Option[o].SetScore(Option[o].ScoreAsOption + Future[f].Score, 1);
  end;
end;

procedure TListOfFutures.ReduceOptions;
var
  o: integer;
  Smaller: double;
begin
  if OptionCount=0 then exit;
  // Get the smaller option score:
  Smaller:= Option[0].ScoreAsOption;
  for o:= 1 to OptionCount-1 do
    Smaller:= Min(Smaller, Option[o].ScoreAsOption);
  // Take Smaller * Sharp frmall scores:
  for o:= 0 to OptionCount-1 do
    Option[o].SetScore(Option[o].ScoreAsOption - Smaller * TPlayer(Owner).Determined, 1);
end;

procedure TListOfFutures.NormalizeOptions;
var
  o: integer;
  Sum: double;
begin
  if OptionCount=0 then exit;
  // Get the sum of scores for all options:
  Sum:= 0;
  for o:= 0 to OptionCount-1 do
    Sum:= Sum + Option[o].ScoreAsOption;
  // Divide option's score by this sum so they sum 1 (normalize):
  if (Sum > 0) then
    for o:= 0 to OptionCount-1 do
      Option[o].SetScore(Option[o].ScoreAsOption / Sum, 1);
end;

// Sum(option * score) will form the "intelligent" decision:
procedure TListOfFutures.AverageOptions(DestinationState: TState);
var
  f: integer;
begin
  // For each free param on the state, use its 2 option's scores to set a value
  for f:= 0 to DestinationState.FreeParamCount-1 do begin
    // Each free param has two options associated, positive and negative:
    DestinationState.FreeParam[f].Value:=
      DestinationState.FreeParam[f].OptionValue * (Option[f*2].ScoreAsOption - Option[f*2+1].ScoreAsOption);
  end;
end;


// *********************
// ** GETTERS/SETTERS **
// *********************

function TListOfFutures.GetFuture(i: integer): TState;
begin
  result:= TState(fFutures[i]);
end;

function TListOfFutures.GetOption(i: integer): TState;
begin
  result:= TState(fOptions[i]);
end;

function TListOfFutures.GetOptionCount: integer;
begin
  result:= High(fOptions) + 1;
end;

function TListOfFutures.GetFutureCount: integer;
begin
  result:= fFutures.Count;
end;


// ********************
// **  DRAWING STUFF **
// ********************

procedure TListOfFutures.DrawOptions;
var
  f: integer;
begin
  for f:= 0 to FutureCount-1 do
    TPlayer(fOwner).DrawFutureScore(Future[f]);
  // Adjust scale of the circles (for the next time):
  TPlayer(fOwner).ResetFutureScoreScale;
end;


end.
