unit UPhysic2D;

// Physic 2D objet with its physic simulation.
// It can represnt a stone, or a rocket, both have gravity, etc.

interface

uses
  ExtCtrls, Graphics, Math, Types, DateUtils, SysUtils, Contnrs, Forms,
  UState, UGoal, UPlayer, UPlayer2D, UUtils;

type
  TPhysic2D = class(TPlayer2D)
  private
    // Gravity? Karts don't use, Rockets do:
    fGravity: double;
    // Impact point with track limits:
    fImpactX, fImpactY: double;
    fLanded: boolean;
    // Accesing player's default goals:
    function  GetGoalRaced: TGoal;
    procedure SetGoalRacedStrength(v: double);
    function  GetGoalRacedStrength: double;
  protected
    // Indexes of the StateParams that are common to all TPlayers:
    fVelX, fVelY, fVelA: integer;
    // Physic properties:
    fDensity, fMass, fAngInertia: double;
    // Accelarations are temporary vars, not real state params:
    fAccX, fAccY, fAccA: double;
    // Indexes of default goals:
    fGoalRaced: integer;
    // Size change -> Inertia invalidated, needs recalc.
    procedure SetSize(v: double); override;
    // Properties getter/setters:
    procedure SetVelX(v: double);
    function  GetVelX: double;
    function  GetOldVelX: double;
    procedure SetVelY(v: double);
    function  GetVelY: double;
    function  GetOldVelY: double;
    procedure SetVelA(v: double);
    function  GetVelA: double;
    function  GetMass: double;
    function  GetAngInertia: double;
    procedure SetDensity(v: double);
    // Simulation stuff:
    procedure PreSimulation; override;
    procedure InSimulation; override;
    procedure PostSimulation; override;
    procedure DefineStateParams; override;
  public
    constructor create(x, y, a: double); virtual;
    // Simulation stuff:
    procedure BounceWithTrack;
    function  CheckCollision(Other: TPlayer): boolean; override;
    // Impact point:
    function IsInsideTrack: boolean; override; // Check for impacts and record impact point
    property ImpactX: double read fImpactX write fImpactX;
    property ImpactY: double read fImpactY write fImpactY;
    property Landed: boolean read fLanded write fLanded;
    // Physic properties:
    property Gravity: double read fGravity write fGravity;
    property Mass: double read GetMass;
    property AngInertia: double read GetAngInertia;
    property Density: double read fDensity write SetDensity;
    // Velocities:
    property VelX:  double read GetVelX write SetVelX;
    property VelY:  double read GetVelY write SetVelY;
    property VelA:  double read GetVelA write SetVelA;
    property OldVelX:  double read GetOldVelX;
    property OldVelY:  double read GetOldVelY;
    // Accelerations:
    property AccX: double read fAccX write fAccX;
    property AccY: double read fAccY write fAccY;
    property AccA: double read fAccA write fAccA;
    // Properties to manipulate the player's goals:
    property GoalRaced: TGoal read GetGoalRaced;
    property GoalRacedStrength: double read GetGoalRacedStrength write SetGoalRacedStrength;
  end;


implementation

uses UTrack, UTrack2D, UGoal2DRaced;


// ****************************
// **  ADMINISTRATIVE STUFF  **
// ****************************

constructor TPhysic2D.create(x, y, a: double);
begin
  inherited create;
  PlaceHere(x, y, a);
  Size:= 8;
  Density:= 2; // A rock!
  // Physic objects have inertia:
  Gravity:=    9.8; // Set to 0 if you dont want to fall down
  fDensity:=     1;
  fMass:=       -1; // -1 = needs recalc.
  fAngInertia:= -1;
  fLanded:= true;
  // 2D players have a natural "tendency" for racing fast and inside the track:
  GoalSet.AddGoal(TGoal2DRaced.create(Self));
  fGoalRaced:= GoalSet.Count-1;
  SetIniState;
end;


// **********************
// ** STATE DEFINITION **
// **********************

procedure TPhysic2D.DefineStateParams;
begin
  inherited;
  State.AddParam(TStateParam.createComputed(State, 'VelX', 0, 0, 0));
  State.AddParam(TStateParam.createComputed(State, 'VelY', 0, 0, 0));
  State.AddParam(TStateParam.createComputed(State, 'VelA', 0, 0, 0));
  // Index of created params:
  fVelX:= State.ParamIndexByName('VelX');
  fVelY:= State.ParamIndexByName('VelY');
  fVelA:= State.ParamIndexByName('VelA');
end;


// *******************
// ** DRAWING STUFF **
// *******************


// ***********************
// **  SIMULATION STUFF **
// ***********************

procedure TPhysic2D.PreSimulation;
begin
  inherited;
  AccX:= 0;
  AccY:= 0;
  AccA:= 0;
  // Leaving the screen is really bad, kill him!
  if not IsInsideImage then
    health:= 0;
end;

procedure TPhysic2D.InSimulation;
var
  Uncertain: double;
begin
  inherited;
  // Reset accelerations:
  AccY:= Gravity;
  // Imagining is not the same thing as really moving:
  if Imagining then begin
    // Apply velocity uncertainty: Your velocity is not accurately imagined.
    if (Uncertainty>0) then begin
      Uncertain:= RandG(0, Uncertainty) * StepSeconds;
      VelX:= VelX + Uncertain;
      VelY:= VelY + Uncertain;
    end;
    // GoalLess players die if some health is lost, and then full stop:
    if GoalLess and Dead then begin
      // Full stop!
      AccY:= 0;
      VelX:= 0;
      VelY:= 0;
      VelA:= 0;
    end;
  end;
end;

// InSimultion calculated accs., now apply them to the object:
procedure TPhysic2D.PostSimulation;
var
  s, s2: double;
begin
  inherited;
  // Goal-Less players must imagine they get stuck if they die:
  if GoalLess and Imagining and (Health<100) then
    exit;
  // Usual way:
  s:=  StepSeconds;
  s2:= sqr(s)/2;
  // Apply some air friction:
  VelX:= VelX * (1-0.01*s);
  VelY:= VelY * (1-0.01*s);
  VelA:= VelA * (1-0.20*s);
  // Apply accelerations:
  VelX:= VelX + AccX * s;
  VelY:= VelY + AccY * s;
  VelA:= VelA + AccA * s;
  // Apply velocity to position:
  // Remember: Pos(t+s) = Pos(t) + Vel(t)*s + 1/2 Acc(t)*s^2
  PosX:=  PosX +  VelX * s + AccX * s2;
  PosY:=  PosY +  VelY * s + AccY * s2;
  Angle:= Angle + VelA * s + AccA * s2;
  // Update nodes of the skeleton:
  NodesNeedRecalc:= true;
end;


// Is the player inside the track, or is it crashing?
function TPhysic2D.IsInsideTrack: boolean;
var
  i: integer;
  nImp: integer;
begin
  CalcNodes;
  NImp:= 0;
  fImpactX:= 0;
  fImpactY:= 0;
  // Ask the Track if the player is on the track or outside it.
  for i:= Low(Nodes) to High(Nodes) do begin
    // If one point is off-track, then player is off-track:
    if not TTrack2d(Track).IsInsideTrack(Nodes[i].x, Nodes[i].y) then begin
      fImpactX:= fImpactX + NodesR[i, 1];
      fImpactY:= fImpactY + NodesR[i, 2];
      inc(NImp);
    end;
  end;
  if (NImp>0) then begin
    // Average impact points:
    fImpactX:= fImpactX / NImp;
    fImpactY:= fImpactY / NImp;
  end else begin
    // Memorize position: "last known inside position"
    MemorizePosition;
  end;
  result:=   (NImp=0);
end;

// Bounce back if crash into black pixels:
procedure TPhysic2D.BounceWithTrack;
var
  i: integer;
  // Vars to react from a impact on a node of the player's body:
  C2I_x, C2I_y, dist: double; // Vector Player's center -> PointOfImpact
  C2In_x, C2In_y: double;     // Same I2C but normalized to length 1.
  RF_x, RF_y, RF_l: double;   // Reaction vel. = -Velocity of the point of impact.
  Cross, Prod: double;
  // Small 1/10 steps to the last inside position:
  ActX, ActY, ActA: double; //  change variable name! And RF is not a force, is a velocity
  // Accumulate all vels. that will be added to player's state: you can not add
  // them on each impact point found as it would interfer with the next impact
  // point you find and will over react to the first impact point:
  AddVelX, AddVelY, AddVelA: double;
  Energy0: double;
  ImpX, ImpY: double;
const
  Elasticity: double = 1.30; // 1.5 - 1.9 works OK
  Debug: boolean = false;
  function KyneticEnergy: double;
  begin
    result:= Mass * (sqr(VelX/100)+sqr(VelY/100)) +
             AngInertia * sqr(VelA/100);
  end;

begin
  if IsInsideTrack then
    exit;
  // We need to control the amount of kynetic energy before and after to avoid
  // random jumps due to pixels vs vectors calculations (E=mv2 but we keep v):
  Energy0:= KyneticEnergy;
  // Walk back in time to the impact time (in 1/10 steps):
  // Note1: If you crashed too hard, you can have more than half of the player
  // outside. Using center point to bounce will work "reversed".
  // Note2: After calculating reactions, a final 1/10 step must be done in order
  // to let the player inside the track.
  ActX:= (State.Param[fPosX ].MemoValue - PosX)  / 10;
  ActY:= (State.Param[fPosY ].MemoValue - PosY)  / 10;
  ActA:= (State.Param[fAngle].MemoValue - Angle) / 10;
  i:= 20; // Better 20 than 10... no more than 9 should be used.
  repeat
    ImpX:= ImpactX;
    ImpY:= ImpactY;
    PosX:=  PosX  + ActX;
    PosY:=  PosY  + ActY;
    Angle:= Angle + ActA;
    dec(i);
  until IsInsideTrack or (i<=0);
  // --- debug ---
  if Debug and not Imagining then begin
    TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
    TTrack2D(Track).Img.Canvas.Pen.Color:= clRed;
    TTrack2D(Track).Img.Canvas.Ellipse(round(ImpactX-2), round(ImpactY-2), round(ImpactX+2), round(ImpactY+2));
    Application.ProcessMessages;
  end;
  // C2I = Vector from players center to impact point:
  C2I_x:= ImpX - PosX;
  C2I_y:= ImpY - PosY;
  // Well, I really will need it normalized to length 1 later:
  dist:= Sqrt(Sqr(C2I_x)+Sqr(C2I_y));
  // Impact point = center point? Not possible, but...
  if (dist<=Size/100) then exit;
  // Normalize it:
  C2In_x:= C2I_x / dist;
  C2In_y:= C2I_y / dist;
  //
  // Real velocity of the impact point: Player's vel. + Rotational vel.
  // Note1: Normal vector used (-C2I_y, +C2I_x), Rot.Vel = Normal * VelRot(rad/s).
  // Note2: The reaction force of the ground is -VReal, so I change sings.
  RF_x:= -(VelX - C2I_y * VelA);
  RF_y:= -(VelY + C2I_x * VelA);
  // --- debug ---
  if Debug and not Imagining then begin
    TTrack2D(Track).Img.Canvas.Pen.Color:= clRed;
    TTrack2D(Track).Img.Canvas.MoveTo(round(ImpX), round(ImpY));
    TTrack2D(Track).Img.Canvas.LineTo(round(ImpX+RF_x), round(ImpY+RF_y));
    Application.ProcessMessages;
  end;
  // RF_L = Damage to the player!
  RF_l:= Sqrt(Sqr(RF_x)+Sqr(RF_y));
  // Change the player's velocity with bounce = FR proyected over C2I:
  Cross:= ((RF_x * C2In_X) + (RF_y * C2In_Y));
  AddVelX:= (Cross * C2In_x);
  AddVelY:= (Cross * C2In_y);
  // --- debug ---
  if Debug and not Imagining then begin
    TTrack2D(Track).Img.Canvas.Pen.Color:= clBlue;
    TTrack2D(Track).Img.Canvas.MoveTo(round(PosX), round(PosY));
    TTrack2D(Track).Img.Canvas.LineTo(round(PosX+AddVelX), round(PosY+AddVelY));
    Application.ProcessMessages;
  end;
  // Rest of the reaction force? (Must be perpendicular to I2C)
  RF_x:= RF_x - (Cross * C2I_x);
  RF_y:= RF_y - (Cross * C2I_y);
  // Length of N (normal velocity) / radio = angular velocity to add:
  // Note: N is the z-coordinte on vectorial product (C2In ^ RF).
  Prod:= (C2In_x * RF_Y) - (C2In_y * RF_X);
  AddVelA:= (Prod*Elasticity / dist);
  // Correct velocity of the player as a reaction to all the impact points found:
  if abs(AddVelX) + abs(AddVelY) > 1 then begin
    VelX:= VelX + Elasticity * AddVelX;
    VelY:= VelY + Elasticity * AddVelY;
    VelA:= VelA + sqrt(Elasticity) * AddVelA;
  end else begin
    VelX:= VelX + 2 * AddVelX;
    VelY:= VelY + 2 * AddVelY;
    VelA:= VelA + 2 * AddVelA;
  end;
  // Control that energy is not created or destroyed:
  while KyneticEnergy > Energy0 do begin
    VelX:= VelX * 0.9;
    VelY:= VelY * 0.9;
    VelA:= VelA * 0.9;
  end;
  // Remember the impact point:
  ImpactX:= ImpX;
  ImpactY:= ImpY;
  // Take health away after crash:
  BouncingDamage(RF_l);
end;

// 2 Players collide when a node of one is inside of the other's poligon.
// Point inside a convex poligon? All cross-products must be same signed.
// All points of A in B or vice versus are averaged to obtain a contact point.
function TPhysic2D.CheckCollision(Other: TPlayer): boolean;
var
  Other2D, P1, P2: TPhysic2D;
  px, py: double;
  pi, i, ni, j: integer;
  NPts: integer;
const
  // Check intermediate points too:
  SubDiv: integer = 5;
  Debug: boolean = false;
begin
  result:= false;
  // Only TPhysic2D object can be checked:
  if (Other is TPhysic2D) then
    Other2D:= TPhysic2D(Other)
  else
    exit;
  // Force node recalculation if necessary:
  CalcNodes;
  Other2D.CalcNodes;
  // If points of A are inside B, calc impact point and return true:
  NPts:= 0;
  ImpactX:= 0;
  ImpactY:= 0;
  // Points of the other inside me? Ohhh!
  for pi:= 1 to 2 do begin
    if (pi=1) then begin
      P1:= Self;
      P2:= Other2D;
    end else begin
      P2:= Self;
      P1:= Other2D;
    end;
    // For each node:
    for i:= 0 to High(P2.Nodes) do begin
      // Index of next node:
      if (i = High(P2.Nodes)) then
        ni:= 0
      else
        ni:= i+1;
      // For each intermediate point (inter-node):
      for j:= 0 to SubDiv-1 do begin
        // Intermediate point:
        Px:= P2.Nodes[i].x + (P2.Nodes[ni].x - P2.Nodes[i].x) * (j/SubDiv);
        Py:= P2.Nodes[i].y + (P2.Nodes[ni].y - P2.Nodes[i].y) * (j/SubDiv);
        // Inside other's body?
        if P1.IsInsidePlayer(Px, Py) then begin
          inc(NPts);
          ImpactX:= ImpactX + Px;
          ImpactY:= ImpactY + Py;
          if Debug and not Imagining then begin
            TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
            TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
            TTrack2D(Track).Img.Canvas.Rectangle(round(Px-1), round(Py-1), round(Px+1), round(Py+1));
          end;
        end;
      end;
    end;
  end;
  // Average intersecting points to get the impact point:
  if (NPts>0) then begin
    result:= true;
    ImpactX:= ImpactX / NPts;
    ImpactY:= ImpactY / NPts;
    Other2D.ImpactX:= ImpactX;
    Other2D.ImpactY:= ImpactY;
  end;
end;


// *****************
// **  GOAL STUFF **
// *****************

function TPhysic2D.GetGoalRaced: TGoal;
begin
  result:= GoalSet.Goal[fGoalRaced];
end;
function TPhysic2D.GetGoalRacedStrength: double;
begin
  result:= GoalRaced.Strength;
end;
procedure TPhysic2D.SetGoalRacedStrength(v: double);
begin
  GoalRaced.Strength:= v;
end;


// ********************
// ** GETTER/SETTERS **
// ********************

// Size (if changed, mass is invalidated):
procedure TPhysic2D.SetSize(v: double);
begin
  // Size must be positive (zero mass is not allowed):
  if (v>0) then begin
    // Only fires it size actually changes:
    if (State.Param[fSize].Value <> v) then begin
      inherited;
      // Mass and RotInertia are invalidated:
      fMass:=       -1;
      fAngInertia:= -1;
    end;
  end;
end;

// VelX:
procedure TPhysic2D.SetVelX(v: double);
begin
  State.Param[fVelX].Value:= v;
end;
function  TPhysic2D.GetVelX: double;
begin
  result:= State.Param[fVelX].Value;
end;
function  TPhysic2D.GetOldVelX: double;
begin
  result:= State.Param[fVelX].OldValue;
end;

// VelY:
procedure TPhysic2D.SetVelY(v: double);
begin
  State.Param[fVelY].Value:= v;
end;
function  TPhysic2D.GetVelY: double;
begin
  result:= State.Param[fVelY].Value;
end;
function  TPhysic2D.GetOldVelY: double;
begin
  result:= State.Param[fVelY].OldValue;
end;

// VelA (angular velocity):
procedure TPhysic2D.SetVelA(v: double);
begin
  State.Param[fVelA].Value:= v;
end;
function  TPhysic2D.GetVelA: double;
begin
  result:= State.Param[fVelA].Value;
end;

// Mass:
function TPhysic2D.GetMass: double;
begin
  if (fMass < 0) then
    fMass:= Density * Power(Size/UnitSize, 3);
  result:= fMass;
end;

// AngInertia:
function TPhysic2D.GetAngInertia: double;
begin
  if (fAngInertia < 0) then
    fAngInertia:= Mass * Power(Size/UnitSize, 2);
  result:= fAngInertia;
end;

// Density (affects mass and angular inertia automatically):
procedure TPhysic2D.SetDensity(v: double);
begin
  fDensity:= v;
  // Invalidate old values:
  fMass:= -1;
  fAngInertia:= -1;
end;


end.
