unit UPlayer;

interface

uses Graphics, Types, Contnrs,
     UState, UListOfFutures, UGoal, UGoalSet;

type
  TPlayer = class(TObject)
  private
    // Team the player belongs to:
    fTeam: TObject;
    // Color for drawing:
    fColor: TColor;
    // Debug flags:
    fDebugTraces: boolean;
    fDebugFeelings: boolean;
    fDebugOptions: boolean;
    fDebugLevels: boolean;
    // Debuging options' score is tricky, some times range from 0 to 1000.
    fFutureScoreScale, fFutureScoreMaxR: extended;
    // Am I imagining futures by taking random decisions?
    // When passing from false to true you must "initialize" FutureState, and
    // when setting it back to false, you must store the future found in your
    // internal list.
    fImagining: boolean;
    // Uncertaintly when imagining your position/velocity?
    fUncertainty: double;
    // How fast you are deciding among similar things?
    fDetermined: double;
    // Set of goals the player follows:
    fGoalSet: TGoalSet;
    // I duplicated my self in the last frame, flag it so goals can use it.
    fHadDuplicated: boolean;
    // Joystick allows one limiter per degree of freedom:
    fUseLimiters: boolean;
    // Flag for dead players:
    fDead: integer; // -1 = not set, look at helath=0 to know.
    fAutoGridSize: boolean;
    fEntropicEffect: double; // 0.1 = 10% of the futures must be discarted and 90% differents
    //
    // ************************
    // **  SETTER & GETTERS  **
    // ************************
    //
    // Actual state of the player (real or imaginary for futures):
    function GetState: TState;
    function GetTrack: TObject;
    // MicroState components:
    // ======================
    function  GetDead: boolean;
    procedure SetEnergy(v: double);
    function  GetEnergy: double;
    function  GetOldEnergy: double;
    function  GetIniEnergy: double;
    procedure SetHealth(v: double);
    function  GetHealth: double;
    function  GetOldHealth: double;
    function  GetIniHealth: double;
    function  GetLimiter(n: integer): double;
    procedure SetLimiter(n: integer; v: double);
    function  GetLimPush(n: integer): double;
    procedure SetLimPush(n: integer; v: double);
    // Grid size control:
    procedure SetGridSize(v: integer);
    // Animated -can decide on some free params- object flag:
    function  GetAnimated: boolean;
    // Accesing player's default goals:
    function  GetGoalEnergy: TGoal;
    procedure SetGoalEnergyStrength(v: double);
    function  GetGoalEnergyStrength: double;
    function  GetGoalHealth: TGoal;
    procedure SetGoalHealthStrength(v: double);
    function  GetGoalHealthStrength: double;
    function  GetGoalLess: boolean;
  protected
    // A "State" contains all parameters (time-changing, fixed, AI driven) the player.
    // There are 2 ones: the actual state of the "real" player, and the "imaginary"
    // state used to imagine a future for this player. I will use "State" in the
    // code, but it will mask one of the two possible states depending on begin
    // really deciding with futures previously found, or you are precisely looking
    // for those futures.
    fActualState, fFutureState: TState;
    // Indexes of the StateParams that are common to all TPlayers:
    fEnergy, fHealth: integer;
    // Indexes of the Goals that are common to all TPlayers:
    fGoalEnergy: integer;
    fGoalHealth: integer;
    //
    // Grid size to convert from micro to macro state.
    // Computed params with a precision will change it as you redefine grid size.
    // If you define a target number of futures needed to decide properly, then
    // "AutoGridSize" will make grid size to change so the number of futures found
    // tend to fit the desired number (thinks better).
    fGridSize, fMaxGridSize, fMinGridSize: integer;
    NodesNeedIni: boolean;
    NodesNeedRecalc: boolean;
    function  DebugThisFuture: boolean; virtual;
    procedure DefineStateParams; virtual;
    function  GetStepMSeconds: integer;
    function  GetFutureMSeconds: integer;
    function  GetStepSeconds: double;
    function  GetCanDuplicate: boolean; virtual;
    procedure SetImagining(v: boolean); virtual;
    property  FutureScoreScale: extended read fFutureScoreScale write fFutureScoreScale;
    property  FutureScoreMaxR:  extended read fFutureScoreMaxR  write fFutureScoreMaxR;
    // Evolution privates:
    function  MakeADuplicate: TPlayer; virtual; abstract;
    // Simulation process is divided into 3 steps:
    // PreSimulation, InSimulation, PostSimulation
    // Other players derived from here will modify it to its own simulation code.
    procedure PreSimulation; virtual;
    procedure InSimulation; virtual;
    procedure PostSimulation; virtual;
    //
  public
    // List of futures found:
    Futures: TListOfFutures;
    //
    constructor create;
    destructor  destroy; override;
    // Back to an initial state:
    procedure ResetState; virtual; //State  = Position, velocity... (change with time)
    // Copy params from another TPlayer:
    procedure CopyFrom(Other: TPlayer); virtual;
    // Drawing stuff:
    procedure Draw; virtual; abstract;
    function  DebugColor(f: TState): TColor; virtual;
    procedure DrawFuturePoint; virtual; abstract;
    // Debugging end point scoring is hard, you need auto scale, so:
    procedure ResetFutureScoreScale;
    procedure DrawFutureScore(f: TState); virtual; abstract; // Level 3 debug
    // Grid size control:
    property GridSize: integer read fGridSize write SetGridSize;
    property AutoGridSize: boolean read fAutoGridSize write fAutoGridSize;
    // After taking a decision, adjust grid size so number of futures found is ok:
    procedure AutoAdjustGridSize; virtual;
    // Save actual state as the intial one:
    procedure SetIniState; virtual;
    property Team:  TObject read fTeam write fTeam;
    property Track: TObject read GetTrack;
    property GoalLess: boolean read GetGoalLess;
    // Debug different things:
    property DebugTraces:   boolean read fDebugTraces   write fDebugTraces;
    property DebugFeelings: boolean read fDebugFeelings write fDebugFeelings;
    property DebugOptions:  boolean read fDebugOptions  write fDebugOptions;
    property DebugLevels:   boolean read fDebugLevels   write fDebugLevels;
    // Color of the player:
    property Color: TColor read FColor write FColor;
    // SIMULATION:
    // ===========
    // Simulate is a closed procedure, it will just call Pre, In and Post simulation
    procedure Simulate;
    property  Animated: boolean read GetAnimated;
    property  StepMSeconds: integer read GetStepMSeconds;
    property  StepSeconds: double read GetStepSeconds;
    property  FutureMSeconds: integer read GetFutureMSeconds;
    function  IsInsideTrack: boolean; virtual; // Check for impacts and record impact point
    function  CheckCollision(Other: TPlayer): boolean; virtual; abstract;
    procedure BouncingDamage(Bounce: double); virtual; abstract;
    // Witch is my state now? (the real or the imaginary one):
    property  State: TState read GetState;
    // Delete all futures as I am going to imagine a new set from my actual state:
    procedure ResetFutures;
    // Start/Stop imagining a new future involves managing the list of futures:
    property  Imagining: boolean read fImagining write SetImagining;
    property  Uncertainty: double read fUncertainty write fUncertainty;
    property  Determined: double read fDetermined write fDetermined;
    // Taking decisions before moving:
    // Take the first random decision of a new future:
    procedure TakeFirstDecision;
    // Take another random decision to continue with a future:
    procedure TakeRandomDecision;
    // Take an intelligent decision based on futures found and their scorings:
    procedure TakeIntelligentDecision;
    // Debug it:
    procedure DrawOptions;
    property UseLimiters: boolean read fUseLimiters write fUseLimiters;
    property EntropicEffect: double read fEntropicEffect write fEntropicEffect;
    property MinGridSize: integer read fMinGridSize write fMinGridSize;
    property MaxGridSize: integer read fMaxGridSize write fMaxGridSize;
    property Dead: boolean read GetDead;
    property Energy: double read GetEnergy write SetEnergy;
    property OldEnergy: double read GetOldEnergy;
    property IniEnergy: double read GetIniEnergy;
    property OldHealth: double read GetOldHealth;
    property IniHealth: double read GetIniHealth;
    property Health: double read GetHealth write SetHealth;
    // Joystick sensitivity control:
    property Limiter[n: integer]: double read GetLimiter write SetLimiter;
    property LimPush[n: integer]: double read GetLimPush write SetLimPush;
    // Properties to manipulate the player's goals:
    property GoalSet: TGoalSet read fGoalSet; // to do work, define GaolSets and use them!
    property GoalEnergy: TGoal read GetGoalEnergy;
    property GoalEnergyStrength: double read GetGoalEnergyStrength write SetGoalEnergyStrength;
    property GoalHealth: TGoal read GetGoalHealth;
    property GoalHealthStrength: double read GetGoalHealthStrength write SetGoalHealthStrength;
    // Evolution:
    function  Clone: TPlayer; virtual;
    property  CanDuplicate: boolean read GetCanDuplicate;
    property  HadDuplicated: boolean read fHadDuplicated write fHadDuplicated;
  end;

implementation

uses Math, SysUtils, UUtils, UTrack, UTeam, UGoalHope, UGoalEnergy, UGoalHealth;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer.create;
begin
  inherited;
  // State:
  fActualState:= TState.create(self);
  fFutureState:= TState.create(self);
  fImagining:= false;
  fDead:= -1;
  // Uncertainty when imaginig:
  Uncertainty:= 0.0; // 0.1 = 10% uncertain for a 1 second long future.
  // Determined intelligence:
  Determined:= 1.0; // <1 = dumb, doesn't decide fast, not optimal.
  // Fill with all the components I need (position, angle, etc):
  fUseLimiters:= true;
  DefineStateParams;
  // Futures:
  Futures:= TListOfFutures.create(self);
  // Grid size control:
  fGridSize:=    10;
  fMinGridSize:= 5;
  fMaxGridSize:= 50;
  fAutoGridSize:= true;
  fEntropicEffect:= 0.5; // Entropic quality: 50% of futures must be discarted
  // Set internal params:
  NodesNeedRecalc:= true;
  NodesNeedIni:= true;
  DebugTraces:= false;
  DebugOptions:= false;
  DebugFeelings:= false;
  DebugLevels:= false;
  FutureScoreScale:= 1;
  fFutureScoreMaxR:= 0;
  // Set a random color:
  Color:= RGB2Color(100+round(random*155), 100+round(random*155), 100+round(random*155));
  SetIniState;
  // Personal goals in life:
  fGoalSet:= TGoalSet.create(self);
  // Default set of goals for all players:
  //
  // Players have HOPE, just being alive is a fixed enjoy.
  //GoalSet.AddGoal(TGoalHope.create(Self));
  // Players have a natural "tendency" for keeping health high:
  GoalSet.AddGoal(TGoalHealth.create(Self));
  fGoalHealth:= GoalSet.Count-1;
  // Players have a natural "tendency" for keeping energy high:
  GoalSet.AddGoal(TGoalEnergy.create(Self));
  fGoalEnergy:= GoalSet.Count-1;
end;

destructor TPlayer.destroy;
begin
  Futures.Free;
  fActualState.Free;
  fFutureState.Free;
  fGoalSet.Free;
  inherited;
end;

procedure TPlayer.CopyFrom(Other: TPlayer);
begin
  // Copy its State:
  fActualState.CopyFrom(Other.fActualState);
  // Reset initial state and scoring:
  SetIniState;
  // Thinking params:
  Uncertainty:=         Other.Uncertainty;
  Self.GridSize:=       Other.GridSize;
  Self.MinGridSize:=    Other.MinGridSize;
  Self.MaxGridSize:=    Other.MaxGridSize;
  Self.AutoGridSize:=   Other.AutoGridSize;
  Self.EntropicEffect:= Other.EntropicEffect;
  Self.UseLimiters:=    Other.UseLimiters;
  Self.DebugTraces:=    Other.DebugTraces;
  Self.DebugOptions:=   Other.DebugOptions;
  Self.DebugFeelings:=  Other.DebugFeelings;
  Self.DebugLevels:=    Other.DebugLevels;
  Self.Imagining:=      Other.Imagining; // Should be false.
end;


// ******************
// ** STATES STUFF **
// ******************

procedure TPlayer.DefineStateParams;
begin
  // Fixed params: (simulation don't change them):
  State.AddParam( TStateParam.createComputed(State, 'Health', 100, 100, 0) );
  State.AddParam( TStateParam.createComputed(State, 'Energy',  50, 100, 0) );
  // Index of all created params (for fast access as property):
  fHealth:=   State.ParamIndexByName('Health');
  fEnergy:=   State.ParamIndexByName('Energy');
end;

// Get actual (real or imaginary) state depending on the proccess you are now:
// Moving or imagining new futures.
function TPlayer.GetState: TState;
begin
  if Imagining then
    result:= fFutureState
  else
    result:= fActualState;
end;

// Reset state to initial:
procedure TPlayer.ResetState;
begin
  NodesNeedRecalc:= true;
  fActualState.BackToIniState;
  GoalSet.Reset;
end;

// Set the actual state as the "initial":
procedure TPlayer.SetIniState;
begin
  // Reset scoring:
  State.ResetScore;
  // Copy actual values into ini and old parts of it.
  State.SetIniState();
end;


// *******************
// ** DRAWING STUFF **
// *******************

// Draw ending points of the futures:
procedure TPlayer.DrawOptions;
begin
  if DebugOptions then
    Futures.DrawOptions;
end;

// Adjust scale of the circles when drawing options/futures
procedure TPlayer.ResetFutureScoreScale;
begin
  // Change scale so ScoreMax => 100 (and use it as a % of the GridSize):
  if (FutureScoreMaxR > 0) then
    FutureScoreScale:= 100 / FutureScoreMaxR;
  // Clean the max:
  fFutureScoreMaxR:= 0;
end;

// Witch color do I use for debugin this future? Red = left, etc.
function TPlayer.DebugColor(f: TState): TColor;
var
  Op: integer;
  R, G, B: integer;
begin
  // Number of the option (except related to joystick):
  if Animated then
    Op:= f.FreeId + f.OptionSign
  else
    Op:= 4;
  // Base color:
  R:= 0; G:= 0; B:= 0;
  case Op of
    0: B:= 255; // Blue = Righ, politically correct
    1: R:= 255; // Red = Left, Viva la resolucion!
    2: begin    // Fucsia = Push forward (usually)
         R:= 255;
         B:= 255;
       end;
    3: begin    // Green = Push backwards:
         R:=  50;
         G:= 255;
       end;
  else
    R:= 160; G:= 160; B:= 160;
  end;
  result:= RGB2Color(R, G, B);
  // Fading the color to white as time goes by?
  //LightColor(result, round(100 * f.TimePassed / TTrack(Track).FutureMSeconds));
end;


// May be you only want some futures to be shown while debugging:
function TPlayer.DebugThisFuture: boolean;
begin
  // By default, plot all futures except if related to joystick sensitivity:
  result:= not Animated or (not State.LimiterOption {and (State.FreeId=0)});
end;


// **********************
// ** SIMULATION STUFF **
// **********************

procedure TPlayer.Simulate;
begin
  PreSimulation;
  InSimulation;
  PostSimulation;
end;

// Previous to simulation the system, some general points are cleared first:
procedure TPlayer.PreSimulation;
var
  i: integer;
begin
  // Take note, time passes by so time passed needs to be updated pre-simultaion
  State.TimePassed:= State.TimePassed + StepMSeconds;
  // Animated players need to adjust joystick sensitivity:
  if Animated then begin
    // Limit sensitivity of joystick while opting:
    if UseLimiters then begin
      for i:= 0 to State.FreeParamCount-1 do
        if Assigned(State.FreeParam[i].Limiter) then
          Limiter[i]:= Limiter[i] + LimPush[i] * StepSeconds;
    end;
  end;
end;

procedure TPlayer.InSimulation;
begin
  // GoalLess players dead if some health is lost:
  if GoalLess and (Health<100) then
    Health:= 0;
  // Each kind of TPlayer will have to do the inner simulation itself.
  // Basically, you will compute accelerations, so the post sim can make the
  // physic evolution.
end;

procedure TPlayer.PostSimulation;
begin
  // A player generally do nothing here, but if you are in a Physic2D player,
  // here we should apply laws of physics to move it usign the forces decided
  // in InSimulation (rockets push by firing thrusters, once pushed, the way
  // it moves is equally valid for other phtsic objects.
end;

// Check for impacts and record impact point
function TPlayer.IsInsideTrack: boolean;
begin
  // By default, player is inside. A Player2D will actually check it:
  result:= true;
end;

// Simulation time (delta time):
function TPlayer.GetStepMSeconds: integer;
begin
  result:= TTrack(Track).StepMSeconds;
end;
function TPlayer.GetStepSeconds: double;
begin
  result:= TTrack(Track).StepSeconds;
end;
function TPlayer.GetFutureMSeconds: integer;
begin
  result:= TTrack(Track).FutureMSeconds;
end;


// *****************
// ** GOALS STUFF **
// *****************


// Accesing player's default goals:
// ================================

function TPlayer.GetGoalEnergy: TGoal;
begin
  result:= GoalSet.Goal[fGoalEnergy];
end;
function TPlayer.GetGoalEnergyStrength: double;
begin
  result:= GoalEnergy.Strength;
end;
procedure TPlayer.SetGoalEnergyStrength(v: double);
begin
  GoalEnergy.Strength:= v;
end;

function TPlayer.GetGoalHealth: TGoal;
begin
  result:= GoalSet.Goal[fGoalHealth];
end;
function TPlayer.GetGoalHealthStrength: double;
begin
  result:= GoalHealth.Strength;
end;
procedure TPlayer.SetGoalHealthStrength(v: double);
begin
  GoalHealth.Strength:= v;
end;

// ********************
// ** THINKING STUFF **
// ********************

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeFirstDecision;
begin
  if Animated then
    State.TakeFirstDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeRandomDecision;
begin
  if Animated then
    State.TakeRandomDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeIntelligentDecision;
begin
  // If you are dead there is nothing to decide, don't waste time!
  if Animated and not dead then begin
    // Analize futures found and take an intelligent decision based on them.
    Futures.TakeIntelligentDecision(State);
    // Auto adjust grid size to tend to a given number of diff. futures.:
    AutoAdjustGridSize;
  end;
end;

// Star/Stop imagining a new future triggers the creation and storing of it
procedure TPlayer.SetImagining(v: boolean);
begin
  // Detect only real changes:
  if (v <> fImagining) then begin
    // Starting to imagine a new future or just coming back to reality?
    if v then begin
      // Start of the imagination proccess of a new future:
      //
      // Basically you only need to set your "imaginary state" as the on/used
      // state as this will be the starting point for the future to come:
      fFutureState.CopyFrom(fActualState);
      // Reset old state:
      fFutureState.SetOldState;
      // This is the initial state for this future:
      fFutureState.SetIniState;
      // Scoring of the future is still empty:
      fFutureState.ResetScore;
      // Flag it if you duplicated during this future:
      HadDuplicated:= false;
      // Reset goal imaginary state:
      GoalSet.Reset;
    end else begin
      // End of the imagination proccess of a new future:
      //
      // The future is found, store it on my list of futures so I can later
      // decide witch was the best decision and take it on the actual state.
      Futures.AddFuture(fFutureState);
    end;
    // Set the new "mind state":
    fImagining:= v;
    // Nodes needs recalc as position has changed:
    NodesNeedRecalc:= true;
    // Dead flag need recalculating:
    fDead:= -1;
  end;
end;

// Delete all futures as I am going to imagine a new set from my actual state:
procedure TPlayer.ResetFutures;
begin
  Futures.Clear;
end;

// Has some free param to decide on?
function TPlayer.GetAnimated: boolean;
begin
  result:= State.Animated;
end;
// Has some goals or is a Goal Less intelligence?
function TPlayer.GetGoalLess: boolean;
begin
  result:= (GoalSet.Count = 0);
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TPlayer.GetCanDuplicate: boolean;
begin
  result:= false;
end;

function TPlayer.Clone: TPlayer;
begin
  // First, duplicate the player:
  result:= MakeADuplicate;
  result.GoalSet.CopyFrom(self.GoalSet);
  HadDuplicated:= true;
  // Add it to my own team:
  TTeam(Team).AddPlayer(result);
  // Evolution tricks:
  if TTrack(Track).Evolution then begin
    // Share your energy with the new baby:
    self.Energy:= self.Energy/2;
    result.Energy:= self.Energy;
    // Change DNA a little (goal stenghts)
    GoalSet.MutateGoals;
    // Use 3 main goal strength to define player's color:
    result.GoalSet.ColorizeOwner;
  end;
end;


// ***********************
// ** GRID SIZE CONTROL **
// ***********************

// Modify the precision of computed params so converting from micro to macro
// state generate more or less different futures. If your player needs a more
// specific GridSize to Precision schema, override it, call inherited, then
// redefine the precisions as needed.
procedure TPlayer.SetGridSize(v: integer);
var i:integer;
begin
  fGridSize:= Min(fMaxGridSize, Max(fMinGridSize, v));
  // By default, all computed params with a precision (those that convert from
  // micro to macro state when a future is reached) copy this "GridSize" as its
  // "Precision", so if only PosX and PosY are defined, a grid size is exactly
  // what you should spect in a vectorial graphic editor.
  for i:= 0 to State.PositionalParamCount-1 do
    State.PositionalParam[i].Precision:= v;
end;

// After taking a decision, adjust grid size so the number of different futures
// tend to a "magical" number. This way, if your are about to crash and you get
// few futures, grid size decreases and futures are more "fine grained" so get
// get more of them and the decision is more stable:
procedure TPlayer.AutoAdjustGridSize;
var
  Discarted: integer;
  Coef: double;
begin
  // If I calculated 100 futures but only 60 where different, then 40% where
  // discarted decrease grid size:
  // Do it by averaging last grid size with estimated one, so it changes smoothly.
  if Animated and AutoGridSize then begin
    // Discarted / Total = Real Entropic effect the player used.
    Discarted:= Max(1, TTrack(Track).NumFutures - Futures.FutureCount);
    Coef:= (EntropicEffect * TTrack(Track).NumFutures) / Discarted;
    GridSize:= Max(MinGridSize, Min(MaxGridSize, round(GridSize * Coef)));
  end;
end;


// *********************
// ** SETTERS/GETTERS **
// *********************

// Dead
function  TPlayer.GetDead: boolean;
begin
  // Define dead flag if not set.
  if fDead = -1 then
    if (Health = 0) then
      fDead:= 1
    else
      fDead:= 0;
  result:= (fDead = 1);
end;

// Energy
procedure TPlayer.SetEnergy(v: double);
begin
  State.Param[fEnergy].Value:= v;
end;
function  TPlayer.GetEnergy: double;
begin
  result:= State.Param[fEnergy].Value;
end;
function  TPlayer.GetOldEnergy: double;
begin
  result:= State.Param[fEnergy].OldValue;
end;
function  TPlayer.GetIniEnergy: double;
begin
  result:= State.Param[fEnergy].IniValue;
end;

// Health = From 100 to 0, if 0 then you are dead!
procedure TPlayer.SetHealth(v: double);
begin
  State.Param[fHealth].Value:= v;
  fDead:= -1;
end;
function  TPlayer.GetHealth: double;
begin
  result:= State.Param[fHealth].Value;
end;
function  TPlayer.GetOldHealth: double;
begin
  result:= State.Param[fHealth].OldValue;
end;
function  TPlayer.GetIniHealth: double;
begin
  result:= State.Param[fHealth].IniValue;
end;

// Joystick sensitivity auto control:

procedure TPlayer.SetLimiter(n: integer; v: double);
begin
  if UseLimiters and Assigned(State.FreeParam[n].Limiter) then
    TStateParam(State.FreeParam[n].Limiter).Value:= v;
end;
function  TPlayer.GetLimiter(n: integer): double;
begin
  if UseLimiters and Assigned(State.FreeParam[n].Limiter) then
    result:= TStateParam(State.FreeParam[n].Limiter).Value
  else
    // Joysticks are defined quite variable, if no auto limiter, then use 1/3
    // of the available spectrum for options (+/-30 -> +/- 10)
    result:= 1/3;
end;

procedure TPlayer.SetLimPush(n: integer; v: double);
begin
  if UseLimiters and Assigned(State.FreeParam[n].LimPush) then
    TStateParam(State.FreeParam[n].LimPush).Value:= v;
end;
function  TPlayer.GetLimPush(n: integer): double;
begin
  if UseLimiters and Assigned(State.FreeParam[n].LimPush) then
    result:= TStateParam(State.FreeParam[n].LimPush).Value
  else
    result:= 0;
end;


function TPlayer.GetTrack: TObject;
begin
  result:= TTeam(fTeam).Track;
end;

end.
