unit UPlayer2D;

// Represent a 2D object with pos and angle, but without any physic simulation,
// no mass, inertia, accelerations or velocity.
// For instance, it can represent a non-physical 2D object like a "zone" of the
// track where there are a different environment (you can recharge energy there)
// Really not very usable, but is an underlaying object under the Physic2D objects.

interface

uses Types,
     UState, UPlayer, UGoal;

type
  TPlayer2D = class(TPlayer)
  private
    // Drawing a future course painting all the points is slooooow:
    fDrawEveryN, fDrawedPoints: integer;
    fUnitSize: double;
    // State components
    // ================
    procedure SetPosX(v: double);
    function  GetPosX: double;
    procedure SetPosY(v: double);
    function  GetPosY: double;
    procedure SetAngle(v: double);
    function  GetAngle: double;
    function  GetOldAngle: double;
    function  GetOldPosX: double;
    function  GetOldPosY: double;
  protected
    // Automatic update of vectors of the body:
    VectorsNeedRecalc: boolean;
    // Indexes of the StateParams that are common to all TPlayers:
    fPosX, fPosY, fAngle, fSize: integer;
    // Simulation code:
    procedure DefineStateParams; override;
    procedure CalcNodes; virtual;
    procedure CalcVectors;
    procedure MemorizePosition;
    // Properties getter/setters:
    procedure SetSize(v: double); virtual;
    function  GetSize: double;
    function  GetOldSize: double;
    function  GetMinSize: double;
    function  GetMaxX: double;
    function  GetMinX: double;
    function  GetMaxY: double;
    function  GetMinY: double;
    // Evolution:
    function  GetCanDuplicate: boolean; override;
  public
    // Those 2D node list forms the body polygon:
    NodesR: array of array[1..2] of double;
    Nodes: array of TPoint;
    // Vectors connecting node i with next node:
    Vectors: array of array[1..2] of double;
    // Administrative:
    constructor create;
    // Change ini and actual 2D position:
    procedure PlaceHere(x, y, a: double);
    // Move to a new relative 2D position:
    procedure HandMove(dx, dy, da: double);
    // Size:
    property Size: double read GetSize write SetSize;
    property OldSize: double read GetOldSize;
    property MinSize: double read GetMinSize;
    property UnitSize: double read fUnitSize write fUnitSize;
    // Simulation:
    procedure BouncingDamage(Bounce: double); override;
    function  IsInsideImage: boolean;
    function  IsInsidePlayer(x, y: double): boolean;
    // Draw the player on the track:
    procedure Draw; override;
    procedure DrawBoxLimits;
    procedure DrawBigPoint;
    // Draw debug information:
    procedure DrawFuturePoint; override;
    procedure DrawFutureScore(f: TState); override;
    // 2D box containg the player:
    property MaxX: double read GetMaxX;
    property MinX: double read GetMinX;
    property MaxY: double read GetMaxY;
    property MinY: double read GetMinY;
    // State params:
    property PosX:    double read GetPosX   write SetPosX;
    property PosY:    double read GetPosY   write SetPosY;
    property Angle:   double read GetAngle  write SetAngle;
    // Previous state params:
    property OldPosX:  double read GetOldPosX;
    property OldPosY:  double read GetOldPosY;
    property OldAngle: double read GetOldAngle;
  end;

implementation


uses Graphics, Math, Forms, SysUtils,
     UUTils, UTrack2D, UTrack;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer2D.create;
begin
  inherited;
  // Set an intial position and angle:
  PosX:= 10;
  PosY:= 10;
  Angle:= 0;
  UnitSize:= 12;
  // 2D player duplicate if evolution is ON:
  //AddGoal(TGoalDuplicate.create(0.5));
  SetIniState;
  // Control if a future needs to pinat 1 point every 5 calculated:
  fDrawEveryN:=   0;
  fDrawedPoints:= 0;
end;

procedure TPlayer2D.DefineStateParams;
begin
  inherited;
  // A 2D TPlayer must have a PosX, PosY Angle and its velocities, at least:
  State.AddParam( TStateParam.createPositional(State, 'PosX',  0.0, 15.0) );
  State.AddParam( TStateParam.createPositional(State, 'PosY',  0.0, 15.0) );
  // Angle is defined as computed so not used in macro states, you could opt differently!
  State.AddParam( TStateParam.createComputed(State, 'Angle', 0.0, 0.0, 0.0) );
  // Size is a fixed param:
  State.AddParam( TStateParam.createFixed(State,    'Size',    6,  25,   3) );
  // Index of all created params (for fast access as property):
  fSize:=  State.ParamIndexByName('Size');
  fPosX:=  State.ParamIndexByName('PosX');
  fPosY:=  State.ParamIndexByName('PosY');
  fAngle:= State.ParamIndexByName('Angle');
end;


// **********************
// ** SIMULATION STUFF **
// **********************

// If player bounce with track limits, calculate damage (-health):
procedure TPlayer2D.BouncingDamage(bounce: double);
var Damage: double;
begin
  // Standar way, bouncing hard is just bad for health:
  // Note: bounce is the velocity change (its length) occurred due to the bounce
  // so the energy associated is m*v^2, but mass is not used! (some day... ).
  if Animated then begin
    Damage:= (Size/UnitSize) * sqr(Bounce/Size) * 10;
    Health:= Health - Damage;
  end;
end;

function TPlayer2D.IsInsideImage: boolean;
var i: integer;
begin
  CalcNodes;
  // Ask the Track if the player's node positions are inside image.
  for i:= Low(Nodes) to High(Nodes) do begin
    if not TTrack2D(Track).IsInsideImage(Nodes[i].x, Nodes[i].y) then begin
      result:= false;
      exit;
    end;
  end;
  result:= true;
end;


// Is the point (x,y) inside the player's polygon?
function TPlayer2D.IsInsidePlayer(x, y: double): boolean;
var
  OK_Side, Side, i: integer;
  Cross: double;
begin
  // From my nodes, build the vectors that connect node i to i+1:
  CalcVectors;
  // Point (x,y) is inside of a convex poligon "B->C->D->E" if and only if all
  // cross-product (B->C)^(B->A) (with B->C a vector of the poligon) has the same
  // sign (i.e. point "A" is always at the right -or left- side of the vectors).
  OK_Side:= 0;
  result:= true;
  for i:= 0 to High(Nodes) do begin
    Cross:= (x - Nodes[i].x) * Vectors[i, 2] - (y - Nodes[i].y) * Vectors[i,1];
    // Point A is in one side of vector B->C of the polygon, remember the side:
    Side:= Sign(Cross);
    // 0 means point A is aligned with vector B->C, just on the border, so
    // don't count it as a prove it is out or in, just forget about this.
    if (Side<>0) then begin
      if (OK_Side=0) then
        // First product non-cero defines the "side" at witch point "A" must
        // always be to be inside the polygon.
        OK_Side:= Side
      else if (OK_Side <> Side) then begin
        // "A" is not at the right side of this vector, so it is outside:
        result:= false;
        break;
      end;
    end;
  end;
end;

// Set the inital "position" manually:
procedure TPlayer2D.PlaceHere(x, y, a: double);
begin
  // Set position:
  PosX:=  x;
  PosY:=  y;
  Angle:= a;
  // Use it as the "initial" position:
  SetIniState;
end;

// Move the player from here:
procedure TPlayer2D.HandMove(dx, dy, da: double);
begin
  // Update my state:
  PosX:=  PosX  + dx;
  PosY:=  PosY  + dy;
  Angle:= Angle + da;
end;

// Take a snapshot of the actual position:
procedure TPlayer2D.MemorizePosition;
begin
  State.Param[fPosX].MemorizeValue;
  State.Param[fPosY].MemorizeValue;
  State.Param[fAngle].MemorizeValue;
end;

// *******************
// ** DRAWING STUFF **
// *******************


// Player drawing:

procedure TPlayer2D.Draw;
var
  C: TCanvas;
  i: integer;
  
  procedure DrawLevel(value: double; desp: integer);
  begin
    C.Pen.Color:= RGB2Color(150,150,150); // Grey
    C.Brush.Color:= clWhite;
    C.Rectangle(round(PosX + Size*1.3 + desp),   round(PosY + Size+1),
                round(PosX + Size*1.3 + desp+4), round(PosY - Size-1));
    // 1=Green to 0=red:
    C.Brush.Color:= RGB2Color(round(255*  Max(0, Min(1, 1.4*(1-value)+0.2))), round(255* Max(0, Min(1, value*1.5-0.25))), 50);
    C.Rectangle(round(PosX+Size*1.3+desp), round(PosY+Size),
                round(PosX+Size*1.3+4+desp), round(PosY+Size-round(2*Size*value/100)));
  end;

  procedure DrawLevel2(value, x1,y1,x2,y2: double; side: integer); //Side 1=left, 2=right
  var
    Sector: array [0..3] of TPoint;
    x3, y3: double;
    Co: TColor;
    BandWidth: double;
  const
    MinWidth: double = 2.5;
  begin
    if (value > 0) then begin
      // 1=Green to 0=red:
      Co:= RGB2Color(round(255*  Max(0, Min(1, 1.4*(1-value)+0.2))), round(255* Max(0, Min(1, value*1.5-0.25))), 50);
      // Get a intermediate point between p1 and p2:
      x3:= round(x1 + (x2-x1)*value);
      y3:= round(y1 + (y2-y1)*value);
      // Define the width of the band to draw:
      BandWidth:= Max(MinWidth, Size*0.1);
      // Fill an exterior -just outside the body- colored band:
      C.Brush.Color:= Co;
      C.Pen.Color:=   Co;
      // Set the 4 points:
      Sector[0].X:= round(x1 + side * Sin(Angle) * BandWidth);
      Sector[0].Y:= round(y1 - side * Cos(Angle) * BandWidth);
      Sector[1].X:= round(x1);
      Sector[1].Y:= round(y1);
      Sector[2].X:= round(x3);
      Sector[2].Y:= round(y3);
      Sector[3].X:= round(x3 + side * Sin(Angle) * BandWidth);
      Sector[3].Y:= round(y3 - side * Cos(Angle) * BandWidth);
      // Draw it:
      C.Polygon(Sector);
    end;
  end;

begin
  inherited;
  if not Assigned(Track) then exit;
  CalcNodes;
  C:= TTrack2D(Track).Img.Canvas;
  // Draw the player's colored body:
  C.Pen.Color:= clBlack;
  C.Brush.Color:= Self.Color;
  C.Polygon(Nodes);
  // Living thing draw more info:
  if Animated then begin
    // Plot health and energy levels:
    i:= floor((High(Nodes))/2);
    DrawLevel2(Health/100, NodesR[i, 1], NodesR[i, 2], NodesR[i-1, 1], NodesR[i-1, 2], -1);
    i:= ceil((High(Nodes)+1)/2);
    DrawLevel2(Energy/100, NodesR[i, 1], NodesR[i, 2], NodesR[i+1, 1], NodesR[i+1, 2], +1);
    // Repaint body lines:
    C.Pen.Color:= clBlack;
    C.Brush.Style:= bsClear;
    C.Polygon(Nodes);
    // Cross dead ones:
    if dead then begin
      C.MoveTo(round(PosX-Size), round(PosY-Size));
      C.LineTo(round(PosX+Size), round(PosY+Size));
      C.MoveTo(round(PosX-Size), round(PosY+Size));
      C.LineTo(round(PosX+Size), round(PosY-Size));
    end;
    if DebugLevels then begin
      // Sensitivity of the joysticks:
      if UseLimiters then begin
        for i:= 0 to round(State.FreeParamCount/2)-1 do
          if State.FreeParam[i*2].Limited then
            DrawLevel(tStateParam(State.FreeParam[i*2].Limiter).Value * 100, i*5);
      end;
    end;
    // Tip: To shows some value as text uncomment:
    // C.Brush.Style:= bsClear;
    // C.TextOut(round(PosX+Size*1.0), round(PosY-2.5*Size), Format('%3.2f', [Value here]) );
  end;
end;

// Draw the 2D bounding box (for debuging):
procedure TPlayer2D.DrawBoxLimits;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
  TTrack2D(Track).Img.Canvas.Pen.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Rectangle(round(MinX), round(MinY), round(MaxX), round(MaxY));
end;
// Draw a big point to remark a position (for debugging):
procedure TPlayer2D.DrawBigPoint;
const
  ra: integer = 2;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsSolid;
  TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
  TTrack2D(Track).Img.Canvas.Brush.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Ellipse(round(PosX-ra), round(PosY-ra), round(PosX+ra), round(PosY+ra));
end;

// Nodes are the points that form the silouette of the player, as a base, I
// will assume it is a triangle, but each descendant of TPLayer must shape
// its nodes to look ok on screen
procedure TPlayer2D.CalcNodes;
var
  i: integer;
const
  Angles: array[0..2] of double = (0.0, 0.8, -0.8);
begin
  // Nodes is a dynamic array, so I must define size on first run:
  if NodesNeedIni then begin
    SetLength(Nodes, 3);
    SetLength(NodesR, 3);
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= Low(Nodes) to High(Nodes) do begin
      NodesR[i,1]:= PosX + Size * cos(Angle + Angles[i]*PI);
      NodesR[i,2]:= PosY + Size * sin(Angle + Angles[i]*PI);
      Nodes[i].X:= round(NodesR[i,1]);
      Nodes[i].Y:= round(NodesR[i,2]);
    end;
    NodesNeedRecalc:=   false;
    VectorsNeedRecalc:= true;
  end;
end;

procedure TPlayer2D.CalcVectors;
var
  NextI, i: integer;
begin
  CalcNodes;
  if VectorsNeedRecalc then begin
    VectorsNeedRecalc:= false;
    // From my nodes, build the vectors that connect node i to i+1:
    SetLength(Vectors, 1+High(Nodes));
    for i:= 0 to High(Nodes) do begin
      if (i=High(Nodes)) then
        NextI:= 0
      else
        NextI:= i+1;
      Vectors[i, 1]:= (NodesR[NextI, 1] - NodesR[i, 1]);
      Vectors[i, 2]:= (NodesR[NextI, 2] - NodesR[i, 2]);
    end;
  end;
end;


// ************************
// ** DRAWING DEBUG INFO **
// ************************


// Draw trace of the future (debug):
procedure TPlayer2D.DrawFuturePoint;
var
  c: TCanvas;
  ra, i: integer;
  Radio: double;
  x0, y0, x1, y1: integer;
  Color1, Color2: TColor;
begin
  inherited;
  // Some futures are hidden at debbug time to avoid meshing too much:
  if (DebugTraces or DebugFeelings) and DebugThisFuture then begin
    // Canvas where to draw:
    c:= TTrack2D(Track).Img.Canvas;
    c.Brush.Style:= bsSolid;
    // Initial and final points (pixels) of the step:
    x0:= round(OldPosX);
    y0:= round(OldPosY);
    x1:= round(PosX);
    y1:= round(PosY);
    Color1:= DebugColor(State);
    // Not so primary color, light it:
    LightColor(Color1, 0.2);
    Color2:= Color1;
    LightColor(Color2, 0.8);
    // Fade to white with time:
    LightColor(Color1, 0.2 * State.TimePassed / FutureMSeconds);
    LightColor(Color2, 0.9 * State.TimePassed / FutureMSeconds);
    // (1) Line conecting points (show loose feelings by turning into black):
    if DebugTraces then begin
      // Darken line color because of loose feelings?:
      if (abs(State.GainCoef-1) < 0.9) then
        DarkColor(Color1, sqr(State.GainCoef));
      // Connect ini and final pos:
      c.Pen.Color:= Color2;
      c.MoveTo(x0, y0);
      c.LineTo(x1, y1);
      //
      // (2) Coloring the pixel based on the option taken:
      // End pixels is bigger:
      if State.TimePassed >= TTrack(Track).FutureMSeconds then begin
        c.Pen.Color:=   DebugColor(State);
        c.Brush.Color:= clBlack;
        c.Rectangle(x1-2, y1-2, x1+1, y1+1);
      end else begin
        c.Pixels[x0, y0]:= Color1;
      end;
      if (fDrawEveryN > 0) then begin
        inc(fDrawedPoints);
        if (fDrawedPoints mod fDrawEveryN = 0) then begin
          fDrawedPoints:= 0;
          c.Pen.Color:= Color1;
          c.Rectangle( round(x1-1), round(y1-1), round(x1+1), round(y1+1) );
        end;
      end;
    end;
    // (3) Mark red/green the points where you loose/gain:
    if Animated and DebugFeelings then begin
      if (abs(State.StepGain-1) > 0.01) then begin
        if State.StepGain < 1 then begin
          Radio:= (2/(1+State.StepGain)-1);
          Color1:= RGB2Color(255,   0,   0);
        end else begin
          Radio:=  sqrt((State.StepGain-1))*5;
          Color1:= RGB2Color(  0, 255,   0);
        end;
        c.Brush.Style:= bsSolid;
        c.Pen.Mode:= pmMask;
        for i:= 1 to 3 do begin
          c.Pen.Color:=   Color1;
          c.Brush.Color:= Color1;
          LightColor(Color1, 0.5);
          ra:= round(Radio * 10 * i/3);
          c.Ellipse(x1-ra, y1-ra, x1+ra, y1+ra);
        end;
      end;
    end;
    c.Pen.Mode:= pmCopy;
    //Application.ProcessMessages;
  end;
end;

// Draw circles around a future ending point with radio = scoring.
// Only the "different" futures of each option will be displayed.
procedure TPlayer2D.DrawFutureScore(f: TState);
var
  x, y, r, mr, n: integer;
  i: integer;
  sr: double;
  a1, a2: double;
  c1, c2: TColor;
  c: TCanvas;
begin
  if f.LimiterOption then exit;
  // Center point of the tile where the future ends:
  x:= f.ParamByName['PosX'].RoundValue;
  y:= f.ParamByName['PosY'].RoundValue;
  // Max radius = half grid size:
  mr:= round(GridSize/2);
  // Size to draw the score as a circular sector inside the tile:
  // I want areas to be proportional to scores, not radius:
  sr:= Min(1, Sqrt(Abs(f.Score)*FutureScoreScale/100));
  r:= round(sr*mr);
  // Store max size for later use in autoscaling.
  FutureScoreMaxR:= Max(FutureScoreMaxR, f.Score);
  // Draw the tile of the grid you have touched (light grey):
  c:= TTrack2D(Track).Img.Canvas;
  c.Pen.Color:= $0AAAAAA;
  c.Pen.Style:= psDot;
  c.Brush.Style:= bsClear;
  c.Rectangle(x-mr, y-mr, x+mr+1, y+mr+1);
  c.Pen.Style:= psSolid;
  // Each free param has it limiter:
  n:= State.FreeParamCount;
  i:= f.FreeId + f.OptionSign;
  // Pie sector starts in a1 and end in a2, in radians:
  a1:= PI + floor(i/2) * 2*PI/n + (i mod 2) * PI - PI/n;
  a2:= a1 + 2*PI/n;
  c1:= DebugColor(f); // Option color
  c2:= c1;
  LightColor(c2, 0.7); // Lighter version
  // Fill part of the tile corresponding to this option (pie sectors):
  c.Pen.Color:=   c1;
  c.Brush.Color:= c2;
  c.Pen.Mode:= pmMask;
  c.Pie(x-r, y-r, x+r, y+r,
    round(x+cos(a1)*50), round(y -sin(a1)*50),
    round(x+cos(a2)*50), round(y -sin(a2)*50) );
  c.Pen.Mode:= pmCopy;
end;


// 2D box containing player:
function  TPlayer2D.GetMaxX: double;
begin
  result:= PosX+Size;
end;
function  TPlayer2D.GetMinX: double;
begin
  result:= PosX-Size;
end;
function  TPlayer2D.GetMaxY: double;
begin
  result:= PosY+Size;
end;
function  TPlayer2D.GetMinY: double;
begin
  result:= PosY-Size;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TPlayer2D.GetCanDuplicate: boolean;
begin
  result:= (Energy>=99);
end;



// *******************
// ** SET/GET STUFF **
// *******************

procedure TPlayer2D.SetPosX(v: double);
begin
  State.Param[fPosX].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosX: double;
begin
  result:= State.Param[fPosX].Value;
end;

procedure TPlayer2D.SetPosY(v: double);
begin
  State.Param[fPosY].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosY: double;
begin
  result:= State.Param[fPosY].Value;
end;

procedure TPlayer2D.SetAngle(v: double);
begin
  // Value is in radians:
  State.Param[fAngle].Value:= v;
  // Force redraw:
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetAngle: double;
begin
  result:= State.Param[fAngle].Value;
end;

// Player's size:
procedure TPlayer2D.SetSize(v: double);
begin
  // Size cant be negative:
  if (v>=0) then begin
    // Only fires it size actually changes:
    if (State.Param[fSize].Value <> v) then begin
      // Record new size:
      State.Param[fSize].Value:= v;
      // Body poligon must be recalculated:
      NodesNeedRecalc:= true;
    end;
  end;
end;

// Size:
function TPlayer2D.GetSize: double;
begin
  result:= State.Param[fSize].Value;
end;
function  TPlayer2D.GetOldSize: double;
begin
  result:= State.Param[fSize].OldValue;
end;
function  TPlayer2D.GetMinSize: double;
begin
  result:= State.Param[fSize].MinValue;
end;


// Old values:
function TPlayer2D.GetOldPosX: double;
begin
  result:= State.Param[fPosX].OldValue;
end;

function  TPlayer2D.GetOldPosY: double;
begin
  result:= State.Param[fPosY].OldValue;
end;

function  TPlayer2D.GetOldAngle: double;
begin
  result:= State.Param[fAngle].OldValue;
end;


end.
