unit URocket;

interface

uses
  ExtCtrls, Graphics, Math, Types, DateUtils, SysUtils, Contnrs, Forms,
  UTrack2D, UState, UPlayer, UPhysic2D, UUtils;

type
  TRocket = class(TPhysic2D)
  private
    // State params' indexes:
    fRotaPower, fMainPower, fRotaPush, fMainPush: integer;
    // State components set/get:
    procedure SetRotaPower(v: double);
    function  GetRotaPower: double;
    procedure SetMainPower(v: double);
    function  GetMainPower: double;
    procedure SetRotaPush(v: double);
    function  GetRotaPush: double;
    procedure SetMainPush(v: double);
    function  GetMainPush: double;
  protected
    // Define your state components:
    procedure DefineStateParams; override;
    // Update the points on screen that form the player's siluette:
    procedure CalcNodes; override;
    // Simulation code:
    procedure InSimulation; override;
    // Evolution:
    function  MakeADuplicate: TPlayer; override;
  public
    constructor create(x, y, a: double); override;
    procedure Draw; override;
    procedure BouncingDamage(Bounce: double); override;
    // State params:
    property  MainPower: double read GetMainPower write SetMainPower;
    property  RotaPower: double read GetRotaPower write SetRotaPower;
    property  MainPush:  double read GetMainPush  write SetMainPush;
    property  RotaPush:  double read GetRotaPush  write SetRotaPush;
  end;


implementation

uses UTrack, UTeam, UGoalLander;


// ****************************
// **  ADMINISTRATIVE STUFF  **
// ****************************

constructor TRocket.create(x, y, a: double);
begin
  inherited;
  PlaceHere(x, y, a);
  Size:= 12;
  Density:= 1.2;
  SetIniState;
  Landed:= false;
  //GoalSet.AddGoal(TGoalLander.create(self));
end;

procedure TRocket.DefineStateParams;
begin
  // PosX and PosY are mandatory, so they are defined on TPLayer2D:
  inherited;
  // Add state params: (Name, Type, Value, Max, Min, Precision)
  // Adjusted for a friction 0.95 and delta time 0.1 (bad physics if it need adjusting):
  State.AddParam(TStateParam.createComputed(State, 'MainPower', 0, 40,   0));
  State.AddParam(TStateParam.createComputed(State, 'RotaPower', 0,  3,  -3));
  State.AddParam(TStateParam.createFree(State, 'MainPush', 100*3));
  State.AddParam(TStateParam.createFree(State, 'RotaPush',  25*3));
  // State param's Indexes:
  fMainPower:= State.ParamIndexByName('MainPower');
  fRotaPower:= State.ParamIndexByName('RotaPower');
  fMainPush:=  State.ParamIndexByName('MainPush');
  fRotaPush:=  State.ParamIndexByName('RotaPush');
end;


// *********************
// **  DRAWING STUFF  **
// *********************

procedure TRocket.CalcNodes;
var
  i: integer;
  r, a: double;
// Polar coordinates for the body points:
const Angl: array[0..4] of double = (0, 26.6, 135, 225, 333);
const Dist: array[0..4] of double = (3, 2.24, 2.83, 2.83, 2.24);
begin
  // Nodes is a dynamic array, so I must set size on first use.
  if NodesNeedIni then begin
    SetLength(Nodes,  1+High(Angl));
    SetLength(NodesR, 1+High(Angl));
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= 0 to High(Angl) do begin
      r:= Dist[i]/3*Size;
      a:= Angl[i]/180*PI;
      NodesR[i, 1]:= PosX + r * cos(Angle + a);
      NodesR[i, 2]:= PosY + r * sin(Angle + a);
      Nodes[i].X:= round(NodesR[i, 1]);
      Nodes[i].Y:= round(NodesR[i, 2]);
    end;
    NodesNeedRecalc:= false;
    VectorsNeedRecalc:= true;
  end;
end;


procedure TRocket.draw;
var
  C: TCanvas;
  IniX, IniY, EndX, EndY: integer;
const WinSize: double = 0.4;
begin
  // TPlayer2D will draw the body for me using the nodes:
  inherited;
  C:= TTrack2D(Track).Img.Canvas;
  // Window for astronuats:
  C.Brush.Color:= clSkyBlue;
  C.Pen.Color:= clBlue;
  C.Ellipse(round(PosX-WinSize*Size), round(PosY-WinSize*Size), round(PosX+WinSize*Size), round(PosY+WinSize*Size));
  // Lower point of the rocket = flames come from there
  IniX:= round( PosX - Cos(Angle) * Size );
  IniY:= round( PosY - Sin(Angle) * Size );
  // Draw main power:
  C.Pen.Color:= clRed;
  EndX:= IniX - round( Cos(Angle) * MainPower );
  EndY:= IniY - round( Sin(Angle) * MainPower );
  C.MoveTo(IniX, IniY);
  C.LineTo(EndX, EndY);
  // Draw left/right rotational power:
  EndX:= IniX + round( cos(Angle+PI/2) * (RotaPower) * 10 );
  EndY:= IniY + round( sin(Angle+PI/2) * (RotaPower) * 10 );
  C.MoveTo(IniX, IniY);
  C.LineTo(EndX, EndY);
end;


// ***********************
// **  SIMULATION STUFF **
// ***********************

// Let run the simulation for a delta time and update state acordingly:
procedure TRocket.InSimulation;
var
  s: double;
  UseEnergy, UseMainPower, UseRotaPower: double;
begin
  inherited;
  // Milliseconds to seconds:
  s:= StepSeconds;
  // When MainPower start, you are not landed anymore:
  if MainPower > 0 then
    Landed:= false;
  // Push joystick => Power
  if Dead or (Energy=0) then begin
    MainPower:= 0;
    RotaPower:= 0;
  end else begin
    MainPower:= MainPower + MainPush * s;
    RotaPower:= RotaPower + RotaPush * s;
  end;
  // Energy used? Not zero even if MainPower is zero... cost of living:
  UseEnergy:= MainPower/40 + 0.01 * sqr(MainPower/4.0);
  Energy:= Max(0, Energy - UseEnergy * s);
  UseEnergy:= Abs(RotaPower/50) + 0.01 * sqr(Abs(RotaPower/5));
  Energy:= Max(0, Energy - UseEnergy * s);
  // The power used is proportional to size: double sized, double power (but not double mass):
  UseMainPower:= MainPower * sqr(Size/UnitSize);
  UseRotaPower:= RotaPower * sqr(Size/UnitSize);
  // Acceleration = Power / Mass
  AccX:= AccX + UseMainPower / Mass * Cos(Angle);
  AccY:= AccY + UseMainPower / Mass * Sin(Angle);
  AccA:= AccA + UseRotaPower / AngInertia;
end;

// If rocket bounce with track limits, calculate damage, but may be it landed OK:
procedure TRocket.BouncingDamage(bounce: double);
var
  Damage, Coef, Absorb, ImpX, ImpY: double;
const
  MinEnergy: double =  90;
  MinHealth: double = 100;
begin
  // Bounce is a velocity, so E = m * v2
  Damage:= (Size/UnitSize) * sqr(Bounce/Size/2) * 15;
  // If contac point is in the rocket basement you can absorb medium impacts and take land:
  ImpX:= (ImpactX-PosX)/Size;
  ImpY:= (ImpactY-PosY)/Size;
  if ((ImpX * Cos(Angle) + ImpY * Sin(Angle)) < -0.4) then begin
    // You landed with the bottom of the rocket down, ok.
    Coef:= 2*Power(max(0, -Sin(Angle)), 1.0);
    Absorb:= (2 + Coef * 5);
  end else begin
    // Ops, you are not upside up, the hull in the upper part is more fragile:
    Absorb:= 2;
  end;
  // Body absorbs part of the damage:
  Damage:= Max(0, Damage - Absorb);
  // Take damage from health:
  if Damage > 0 then
    Health:= Health - Damage;
  // Landed? You need not to be to sloped, almost horizontal.
  if (not dead) and (Health>0) and (Sin(Angle) < -0.80) and
    // Both lower edges must be landed:
    (not TTrack2D(Track).IsInsideTrack(Nodes[2].X, Nodes[2].Y + 2)) and
    (not TTrack2D(Track).IsInsideTrack(Nodes[3].X, Nodes[3].Y + 2)) then begin
    Landed:= true;
    // Recharge?
    if (not TTrack(Track).Evolution) then begin
      // Refill energy and get full health by being repared on land.
      if (Energy < MinEnergy) or (Health < MinHealth) then begin
        // Switch off engine before re-fuelling:
        MainPower:= 0;
        RotaPower:= 0;
        // Recharge energy and health:
        if (Energy < MinEnergy) then
          //Energy:= MinEnergy;
          Energy:= Energy + 20 * StepSeconds;
        if (Health < MinHealth) then
          //Health:= MinHealth;
          Health:= Health + 20 * StepSeconds;
      end else begin
        MainPower:= Max(50 , MainPower); // Give it extra impulse to take off safely
        MainPush:= State.Param[fMainPush].MaxValue;
      end;
    end;
  end;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TRocket.MakeADuplicate: TPlayer;
begin
  // Duplicate yourself:
  result:= TRocket.create(0, 0, 0);
  TRocket(result).CopyFrom(self);
end;

// *******************
// ** SET/GET STUFF **
// *******************

procedure TRocket.SetMainPower(v: double);
begin
  State.Param[fMainPower].Value:= v;
end;
function  TRocket.GetMainPower: double;
begin
  result:= State.Param[fMainPower].Value;
end;

procedure TRocket.SetRotaPower(v: double);
begin
  State.Param[fRotaPower].Value:= v;
end;
function  TRocket.GetRotaPower: double;
begin
  result:= State.Param[fRotaPower].Value;
end;


procedure TRocket.SetMainPush(v: double);
begin
  State.Param[fMainPush].Value:= v;
end;
function  TRocket.GetMainPush: double;
begin
  result:= State.Param[fMainPush].Value;
end;

procedure TRocket.SetRotaPush(v: double);
begin
  State.Param[fRotaPush].Value:= v;
end;
function  TRocket.GetRotaPush: double;
begin
  result:= State.Param[fRotaPush].Value;
end;

end.
