unit UState;

// A TPlayer always has an "State" indicating it's actual position, velocity, angle, etc.
// A state is a compound of "StateParams" (position.x is one).

// StateParams can be of several types:
//
//   1) ptFixed = Fixed params, it's value won't change with time, like Size,
//      Color, MaxPower, etc. Usually you set them on creation and from
//      there you just read them as you need.
//   2) ptComputed = The values change with time by the simulation itself,
//      like the car position or its velocity.
//   3) ptPositional = PosX is computed, but it is special as its value is used
//      when deciding if two futures ends in the same point.
//   4) ptFree: Correspond to a "degree of freedom", it's value is decided
//      by the AI, like the trottle and the driving wheel position in a kart.
//      Think of them as the "channels" in a RC model, or how many directions
//      you can move the joystick in the RC controler.
//      A RC car has 2 of them (accelerator/turning) while and helicopter can
//      have 3 to 5 depending on models.
//
//      Example: A kart has a computed param called "Turn" that reflects the
//      actual angle of the front tires (or the position of the driving wheel).
//      The AI can't decide the "new position" of the driving wheel, what it
//      decides is the "Force" it will apply to the driving wheel to make it
//      change. This "Turning push" is the "free" param. that AI will fill with
//      the "intelligent" value, and the simultation will use it as a force to
//      change the turning wheel position. Think of it like this: When you drive
//      your car you push left or right with a given force, and push accelator
//      harder or softer, but you don't directly "place" the driving wheel or the
//      acceleator in a decided position.
//
//      Note: If a player has no "Free" params then it is a dumb player, like a
//      rock rolling down the hill, that you need to simulate but it won't
//      take any "decision".
//
// Each StateParam will has a "Precision" that must be interpeted on positional
// ones as a "Grid Size" that makes a point to snap to a imaginary grid.
//
// Also, to avoid the need of storing the "old" and "initial" values in a copy
// of the "State", each param can keep a copy of its old value so you can get
// "deltas" and use them to scrore the changes as you move on into a future.
//
// The initial values are mainly used to reset the simulation to the starting
// point in the user interface, but also during the thinking proccess:
// Once you get 100 future end points for a player, you have to group them
// so all futures with the same initial decision are together. Only inside
// those groups of futures you can discard similar ending points. So the initial
// decision (values of "Free" params for the first frame in the thinking
// proccess) need to be "discretized" so you only have a few choices and safely
// stored in the State so you can retrieve in the decision-making phase,
// so the FreeParam.IniValue will be set on the first move and not changed
// during the rest of the future imagining proccess.

interface

uses Contnrs, Variants;

type TParamType = ( ptFixed, ptComputed, ptPositional, ptFree );

// Each component on the player's state is a TStateParam: PosX, Angle, VelX, etc.
type TStateParam = class(TObject)
  private
    // The name for this param: "PosX", "Heading", "VelX", etc.
    fName: string;
    // The state that owns this StateParam:
    fOwner: TObject;
    // The value itself: angle, coordinate, etc.
    fValue: double;
    // The "old" value (in previous frame) so you can compare and get deltas:
    fOldValue: double;
    // The "initial" value is used in two cases:
    // 1) If you want to "reset" your player to the initial position, you need
    //    to revert all StateParam to theirs ini values.
    // 2) In the case of a param that is a "degree of freedom" (AI can decide
    //    to push it up or down, like the driving wheel angle in a RC Car) and
    //    you are computing a possible future, you need to keep this initial
    //    "decision" as it correspond to a possible "option".
    // Internally, we use integer and round(Value*1000) so comparations work ok.
    fIniValue: integer;
    // Some times you want to memorize a value for later use, like M+ in a
    // calculator. For instance, to bounce back to track it is nice to have
    // the "last in-track known position" to move in that direction the player
    // to avoid keeping outside track after a bounce calculation.
    fMemoValue: double;
    // There are 3 kind of components as stated above: Params, Computed by the
    // simulation and Free (degree of freedom):
    fParamType: TParamType;
    // Boundaries of allowed values for that param (both zero for unbounded):
    fMaxValue, fMinValue: double;
    // You can take random values with a flat or a Gaussian distribution:
    fUseGauss: boolean;
    // Limiters of free params joystick (its index as FreeParams):
    fLimiterId, fLimPushId: integer;
    // Limiter FreeParam can't be limited by itself!
    fDoNotLimit: boolean;
    // Precision when converted into a micro-state (use 0 to discard it):
    // On free params, it is the "small option" or standard deviation for random.
    fPrecision: double;
    // Integer version of the Value so comparing is secure and fast.
    fRoundValueIsValid: boolean;
    fRoundValue: integer;
    procedure SetValue(v: double);
    procedure SetIniValue;
    function  GetIniValue: double;
    function  GetRoundValue: integer;
    procedure SetName(v: string);
    function  GetLimited: boolean;
    function  GetLimiter: TObject;
    function  GetLimPush: TObject;
    function  GetOptionValue: double;
  public
    // Constructor:
    constructor create(Obj: TObject; vName: string; vType: TParamType; vValue, vMax, vMin, vPrec: double); overload;
    constructor createFixed(Obj: TObject; vName: string; vValue, vMax, vMin: double);
    constructor createComputed(Obj: TObject; vName: string; vValue, vMax, vMin: double);
    constructor createPositional(Obj: TObject; vName: string; vValue, vPrec: double);
    constructor createFree(Obj: TObject; vName: string; vMax: double);
    // Before copying form another state I need to create empty StateParams:
    constructor create; overload;
    // Copy from another param:
    procedure CopyFrom(Other: TStateParam);
    // Memorize actual value for later use:
    procedure MemorizeValue;
    // Is it involved in decision making or macro converting?
    function IsFreeParam: boolean;
    function IsPositionalParam: boolean;
    // Limiters associated to a free param:
    property LimiterId: integer read fLimiterId write fLimiterId;
    property Limited: boolean read GetLimited;
    property Limiter: TObject read GetLimiter;
    property LimPush: TObject read GetLimPush;
    // Take a random decision (initial one is different form others):
    function  TakeFirstDecision: integer;
    procedure TakeRandomDecision;
    // Comparing initial decision is better with integers:
    function  IniValueAsInteger: double;
    // Public porperties:
    property Name: string read FName write SetName;
    property Value:      double  read fValue    write SetValue;
    property OldValue:   double  read fOldValue write fOldValue;
    property IniValue:   double  read GetIniValue;
    property MemoValue:  double  read fMemoValue write fMemoValue;
    property RoundValue: integer read GetRoundValue;
    property RoundValueIsValid: boolean read fRoundValueIsValid write fRoundValueIsValid;
    property ParamType: TParamType read FParamType;
    property UseGauss: boolean read fUseGauss write fUseGauss;
    property DoNotLimit: boolean read fDoNotLimit write fDoNotLimit;
    property Precision: double read fPrecision write fPrecision;
    property OptionValue: double read GetOptionValue;
    property Owner: TObject read fOwner;
    property MaxValue:  double read fMaxValue  write fMaxValue;
    property MinValue:  double read fMinValue  write fMinValue;
  end;


// All values of a player's state (the StateParams) form it's TState:
type TState = class(TObject)
  private
    // The TPlayer this state belongs to:
    fOwner: TObject;
    // General list of all its params of any type:
    fParams: TObjectList;
    // Sublist with only free params (to speed up)
    fFreeParams: TObjectList;
    // Sublist with Positional (micro-to-macro) params:
    fPositionalParams: TObjectList;
    // A State can be a future, so have an option number, a free param and a sing:
    fFreeId: integer;         // Free param of the option
    fOptionId: integer;       // Identifier of the option
    fOptionSign: integer;     // Decided pos or negative OptionValue?
    fLimiterOption: boolean;  // Is related to a joystick limiter?
    // Use gaussian distribution on random decision for all params:
    fUseGauss: boolean;
    //
    // Scoring system:
    // Enjoy points accumulated during the future construction (step by step):
    fEnjoy: double;
    // Scoring accumulated on the actual step goal by goal:
    fStepEnjoy, fStepReduc, fStepGain: double;
    // A future can have a time dependent global reduction (if it dies, for instance)
    fGainCoef: double;
    // A final of future sate can be reached by several futures (not in use):
    fHitCount: integer;
    // How long have this future lasted (milliseconds)?
    fTimePassed: int64;
    //
    // Get State param by index or name:
    function GetParam(i: integer): TStateParam;
    function GetFreeParam(i: integer): TStateParam;
    function GetPositionalParam(i: integer): TStateParam;
    function GetAnimated: boolean;
    function GetParamByName(v: string): TStateParam;
    procedure SetUseGauss(v: boolean);
    function GetScore: double;
    function GetScoreAsOption: double;
  public
    // Create and destroy it:
    constructor create(MyOwner: TObject);
    destructor  destroy; override;
    // Add another StateParam:
    procedure AddParam(v: TStateParam);
    // Copy values form another TState:
    procedure CopyFrom(Other: TState);
    // Number of params of this state:
    function ParamCount: integer;
    function FreeParamCount: integer;
    function PositionalParamCount: integer;
    // Debug only:
    function  GetDecisionVector: string;
    // Set the OldValues to Value.
    procedure SetOldState;
    // Go back to the previous or ini state:
    procedure BackToOldState;
    procedure BackToIniState;
    // Take actual state as your "initial" -and thereof your "old"- state:
    procedure SetIniState;
    // Take a random decision (initail one is different form others):
    procedure TakeFirstDecision;
    procedure TakeRandomDecision;
    // Future comparation stuff:
    function HaveSameOption(Other: TState): boolean;
    function HaveSameFinalState(Other: TState): boolean;
    // Adding a goal scoring to my step enjoy:
    procedure ResetStepScoring;
    procedure ResetScore;
    procedure AddGoalScoring(Goal: TObject);
    procedure AddStepScoring;
    // Player that owns this state:
    property Owner: TObject read fOwner;
    //property BelongsTo: TObject read fBelongsTo write fBelongsTo;
    property OptionId: integer read fOptionId write fOptionId;
    property OptionSign: integer read fOptionSign write fOptionSign;
    property FreeId: integer read fFreeId write fFreeId;
    property LimiterOption: boolean read fLimiterOption write fLimiterOption;
    property Param[i: integer]: TStateParam read GetParam;
    property FreeParam[i: integer]: TStateParam read GetFreeParam;
    property PositionalParam[i: integer]: TStateParam read GetPositionalParam;
    // Easy access to a param by its name:
    function ParamIndexByName(v: string): integer;
    property ParamByName[n: string]: TStateParam read GetParamByName;
    // How much does this future scored?
    procedure SetScore(e, g: double); // dangerous, do not attach to score property!
    procedure CopyScoreFrom(f: TState);
    property Score: double read GetScore;
    property ScoreAsOption: double read GetScoreAsOption;
    property GainCoef: double read fGainCoef write fGainCoef;
    property StepGain: double read fStepGain write fStepGain;
    property Enjoy: double read fEnjoy write fEnjoy;
    property HitCount: integer read fHitCount write fHitCount;
    property TimePassed: int64 read fTimePassed write fTimePassed;
    property UseGauss: boolean read fUseGauss write SetUseGauss;
    // Does this player have any free param to decide on?
    property Animated: boolean read GetAnimated;
  end;


implementation

uses UPlayer, Math, SysUtils, Classes, UGoal;


// *****************
// ** TStateParam **
// *****************

// **************************
// **     TStateParam      **
// ** ADMINISTRATIVE STUFF **
// **************************

// Complete create with all values set on origin:
constructor TStateParam.create(Obj: TObject; vName: string; vType: TParamType; vValue, vMax, vMin, vPrec: double);
begin
  inherited create;
  // Initialize:
  fOwner:= Obj;
  Name:=       vName;
  fParamType:= vType;
  fMaxValue:=  vMax;
  fMinValue:=  vMin;
  fDoNotLimit:=false;
  Value:=      vValue;
  OldValue:=   vValue;
  MemoValue:=  vValue;
  fLimiterId:= -1;
  fLimPushId:= -1;
  SetIniValue;
  fUseGauss:= true; // "false" is worst! Random decision will be too narrow or too wide.
  fPrecision:= vPrec;
end;

constructor TStateParam.createFixed(Obj: TObject; vName: string; vValue, vMax, vMin: double);
begin
  create(Obj, vName, ptFixed, vValue, vMax, vMin, 0);
end;
constructor TStateParam.createComputed(Obj: TObject; vName: string; vValue, vMax, vMin: double);
begin
  create(Obj, vName, ptComputed, vValue, vMax, vMin, 0);
end;
constructor TStateParam.createPositional(Obj: TObject; vName: string; vValue, vPrec: double);
begin
  create(Obj, vName, ptPositional, vValue, 0, 0, vPrec);
end;
constructor TStateParam.createFree(Obj: TObject; vName: string; vMax: double);
begin
  create(Obj, vName, ptFree, 0, vMax, -vMax, 0);
end;

// Fake create to get an empty StateParam that needs to be "copyied from"
constructor TStateParam.create;
begin
  create(nil, '', ptFixed, 0.0, 0.0, 0.0, 0.0);
end;

procedure TStateParam.SetValue(v: double);
begin
  // Boundaries are active when they "make sense" (two zeros doesn't make sense)
  if (fMaxValue > fMinValue) then
    // Avoid out of boundaries values:
    fValue:= Min(fMaxValue, Max(fMinValue, v))
  else
    fValue:= v;
  // Rounded value need to be recalc if needed:
  RoundValueIsValid:= false;
end;

function TStateParam.GetRoundValue: integer;
begin
  if not RoundValueIsValid then begin
    if (Precision = 0) then begin
      fRoundValue:= round(Value);
    end else begin
      fRoundValue:= round(Precision * round(Value/Precision));
    end;
    RoundValueIsValid:= true;
  end;
  result:= fRoundValue;
end;


// Use actual value as the IniValue, rounding it first (ini values are rounded
// for better comparation) but keep 4 decimals:
procedure TStateParam.SetIniValue;
begin
  fIniValue:= round(1000*Value);
end;

function TStateParam.GetIniValue: double;
begin
  result:= fIniValue / 1000;
end;
function TStateParam.IniValueAsInteger: double;
begin
  result:= fIniValue;
end;

// Names are stored in upper case and trimmed:
procedure TStateParam.SetName(v: string);
begin
  fName:= UpperCase(Trim(v));
end;

// Copy from another param:
procedure TStateParam.CopyFrom(Other: TStateParam);
begin
  fName:=       Other.Name;
  fParamType:=  Other.FParamType;
  fValue:=      Other.fValue;
  fOldValue:=   Other.fOldValue;
  fMemoValue:=  Other.MemoValue;
  fPrecision:=  Other.Precision;
  fIniValue:=   Other.fIniValue;
  fMaxValue:=   Other.MaxValue;
  fMinValue:=   Other.MinValue;
  fUseGauss:=   Other.UseGauss;
  fDoNotLimit:= Other.DoNotLimit;
  fLimiterId:=  Other.LimiterId;
  fLimPushId:=  Other.fLimPushId;
end;

// Memorize actual cvalue for later use:
procedure TStateParam.MemorizeValue;
begin
  fMemoValue:= Value;
end;

// Is it involved in decision making?
function TStateParam.IsFreeParam: boolean;
begin
  result:= (ParamType = ptFree);
end;
// Is it involved in macro state (Positional)?
function TStateParam.IsPositionalParam: boolean;
begin
  result:= (ParamType = ptPositional);
end;

// ********************
// **  TStateParam   **
// ** THINKING STUFF **
// ********************

// Take the initial random decision to start with a new future:
// First decision must be discretized and stored for later analysis.
function TStateParam.TakeFirstDecision: integer;
begin
  // First decisions must be "discretized", so i choose between +/- values:
  result:= floor(2*random);
  case result of
    0: Value:= +OptionValue;
    1: Value:= -OptionValue;
  end;
  // Store as IniValue (the initial decision of the future):
  SetIniValue;
end;

// Take another random decision:
procedure TStateParam.TakeRandomDecision;
var
  UseLimiter: double;
begin
  // Use Gaussian distribution randomness or a pure flat random?
  // A free param representing the LimPush (change in sensibility) do not have
  // to change so much during a future or it will spread too much!
  if Limited then
    UseLimiter:= TStateParam(Limiter).Value
  else
    UseLimiter:= 1/3;
  if UseGauss then begin
    // Using a gaussian bell distribution is recommended:
    Value:= RandG(0, OptionValue) * UseLimiter;
  end else begin
    // Using a flat distribution is ugly and works ugly:
    Value:= (2*random-1) * OptionValue * UseLimiter;
  end;
end;

function TStateParam.GetOptionValue: double;
begin
  if Limited then
    result:= MaxValue * TStateParam(Limiter).Value
  else
    result:= MaxValue;
end;

function  TStateParam.GetLimited: boolean;
begin
  result:= (fLimiterId <> -1);
end;

function  TStateParam.GetLimiter: TObject;
begin
  if Limited then
    result:= TStateParam(TState(Owner).Param[fLimiterId])
  else
    result:= nil;
end;

function  TStateParam.GetLimPush: TObject;
begin
  if Limited then
    result:= TStateParam(TState(Owner).Param[fLimPushId])
  else
    result:= nil;
end;



// ************
// ** TState **
// ************

constructor TState.create(MyOwner: TObject);
begin
  // Create the base object:
  inherited create;
  // Initialize param lists:
  fParams:= TObjectList.create;
  fParams.OwnsObjects:= false; // I create it, I kill it
  // Sub-list with only free params:
  fFreeParams:= TObjectList.create;
  fFreeParams.OwnsObjects:= false;
  // Sub-list with only Positional params:
  fPositionalParams:= TObjectList.create;
  fPositionalParams.OwnsObjects:= false;
  //
  // Remeber the owner:
  fOwner:= MyOwner;
  // Use flat or gaussian distribution for random choices?
  fUseGauss:= true;
  ResetScore;
end;

destructor TState.destroy;
begin
  // Free all the StateParams:
  while fParams.Count > 0 do begin
    TStateParam(fParams.Items[0]).Free;
    fParams.Delete(0);
  end;
  fParams.Free;
  fFreeParams.Free;
  fPositionalParams.Free;
end;

procedure TState.AddParam(v: TStateParam);
var
 Tmp: TStateParam;
begin
  v.fOwner:= self;
  fParams.Add(v);
  // Is it a free param?
  if v.IsFreeParam then begin
    fFreeParams.Add(v);
    // Limiter and LimPush associated (joysticks):
    Tmp:= TStateParam.createComputed(Self, v.Name+'_Limiter', 1/3, 1, 0);
    fParams.Add(Tmp);
    v.fLimiterId:= ParamCount-1;
    Tmp:= TStateParam.createFree(Self, v.Name+'_LimPush', 0.25);
    fParams.Add(Tmp);
    v.fLimPushId:= ParamCount-1;
    fFreeParams.Add(Tmp);
  end;
  // Is it Positional = related to micro to macro?
  if v.IsPositionalParam then
    fPositionalParams.Add(v);
  // If State use Gaussian (Normal) distributions, so you do:
  v.UseGauss:= self.UseGauss;
end;

function TState.GetParam(i: integer): TStateParam;
begin
  if ((i < 0) or (i> ParamCount-1)) then
    raise Exception.create('TState: Param index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fParams[i]);
end;

function TState.GetFreeParam(i: integer): TStateParam;
begin
  if ((i < 0) or (i > FreeParamCount-1)) then
    raise Exception.create('TState: FreeParam index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fFreeParams[i]);
end;

function TState.GetPositionalParam(i: integer): TStateParam;
begin
  if ((i<0) or (i>fPositionalParams.Count-1)) then
    raise Exception.create('TState: Positional Param index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fPositionalParams[i]);
end;

function TState.GetAnimated: boolean;
begin
  // PushLimiter always exists for joystick sensitivity, so you need 2 to animate
  result:= (FreeParamCount > 0);
end;

// Witch index correspond to a param with this name?
function TState.ParamIndexByName(v: string): integer;
var i: integer;
begin
  v:= UpperCase(Trim(v));
  for i:= 0 to ParamCount-1 do begin
    if (v = Param[i].Name) then begin
      result:= i;
      exit;
    end;
  end;
  raise Exception.create('TState: Param with name "'+v+'" not found.');
end;

// This way to access params is easy but slow, keep in mind:
function TState.GetParamByName(v: string): TStateParam;
begin
  result:= Param[ParamIndexByName(v)];
end;


// Copy values form another TState:
procedure TState.CopyFrom(Other: TState);
var i: integer;
begin
  // Both states need to have the same number of params:
  while (ParamCount > Other.ParamCount) do
    fParams.Delete(fParams.Count-1);
  while (ParamCount < Other.ParamCount) do
    AddParam(TStateParam.create());
  // Clear free params' list:
  fFreeParams.Clear;
  fPositionalParams.Clear;
  // Copy all param's info:
  for i:= 0 to ParamCount-1 do begin
    Param[i].CopyFrom(Other.Param[i]);
    if Param[i].IsFreeParam then
      fFreeParams.Add(Param[i]);
    if Param[i].IsPositionalParam then
      fPositionalParams.Add(Param[i]);
  end;
  // Other info:
  UseGauss:=   Other.UseGauss;
  OptionId:=   Other.OptionId;
  OptionSign:= Other.OptionSign;
  LimiterOption:= Other.LimiterOption;
  FreeId:=     Other.FreeId;
  // Copy its scoring:
  Enjoy:=      Other.Enjoy;
  GainCoef:=   Other.GainCoef;
  HitCount:=   Other.HitCount;
end;

// Reset the OldValue in all components of this state.
// Why? After a big state change (you manually move the thing) keeping the old
// state is just dangerous, as it is no longer true.
procedure TState.SetOldState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].OldValue:= Param[i].Value;
end;

// Number of params of this state:
function TState.ParamCount: integer;
begin
  result:= fParams.Count;
end;

// Number of FREE params of this state:
function TState.FreeParamCount: integer;
begin
  result:= fFreeParams.Count
end;

// Number of Positional (Macro related) params of this state:
function TState.PositionalParamCount: integer;
begin
  result:= fPositionalParams.Count
end;

// Go back to the previous state:
// Why? You need to fine-adjust actual position without disturbing the old state,
// for instance, 2 players crash into each other and you want to "walk back" in
// time a little to the exact moment where contact occurred.
procedure TState.BackToOldState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].Value:= Param[i].OldValue;
end;

// Go back to the initial state (reset simulation):
procedure TState.BackToIniState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].Value:= Param[i].IniValue;
end;

// Set actual state as the "initial" state of the player.
procedure TState.SetIniState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do 
    Param[i].SetIniValue;
  SetOldState;
  ResetScore;
end;

// Take first random decision of a future (only set one of the freeparams):
procedure TState.TakeFirstDecision;
begin
  // No free params? Then it can't decide anything, don't waste time:
  if Animated then begin
    // I only decide on ONE 1 free param:
    FreeId:= floor(random * FreeParamCount);
    // Ask it to take a decision:
    OptionSign:= FreeParam[FreeId].TakeFirstDecision;
    // This is my option index to later group easily:
    OptionId:= 2*FreeId + OptionSign;
    // Flag as realted to limiter:
    LimiterOption:= (FreeId mod 2 = 1);
  end;
end;

// Take a random decision on all free params:
procedure TState.TakeRandomDecision;
var i: integer;
begin
  if Animated then
    for i:= 0 to FreeParamCount-1 do
      FreeParam[i].TakeRandomDecision;
end;

procedure TState.SetUseGauss(v: boolean);
var i: integer;
begin
  fUseGauss:= v;
  for i:= 0 to FreeParamCount-1 do
    FreeParam[i].UseGauss:= v;
end;


// **********************
// ** SCORING A FUTURE **
// **********************

// A usual state (a real one or a imaginary one) use this to score it self:
function TState.GetScore: double;
begin
  if TPlayer(Owner).GoalLess then
    result:= 1
  else
    result:= Enjoy * GainCoef;
end;
// A state representing an option need to access the real scoring in goal-less cases:
function TState.GetScoreAsOption: double;
begin
  result:= Enjoy * GainCoef;
end;


procedure TState.CopyScoreFrom(f: TState);
begin
  SetScore(f.Enjoy, f.GainCoef);
end;

procedure TState.SetScore(e, g: double);
begin
  Enjoy:= e;
  GainCoef:= g;
end;

procedure TState.ResetScore;
begin
  Enjoy:= 0;
  HitCount:= 0;
  TimePassed:= 0;
  ResetStepScoring;
  fGainCoef:= 1;
end;

procedure TState.ResetStepScoring;
begin
  fStepEnjoy:= 0;
  fStepReduc:= 1;
  fStepGain:=  1;
end;

// Add the goal's scorings to the step points:
procedure TState.AddGoalScoring(Goal: TObject);
var
  Warning, TimeCoef, Gain: double;
begin
  // Add the enjoy scoring from the goal:
  if (TGoal(Goal).Enjoy > 0) then begin
    // positives correspond to something you enjoy:
    fStepEnjoy:= fStepEnjoy + Sqr(TGoal(Goal).Enjoy) * TGoal(Goal).Strength
  end;
  // Also accumulate (product) those reduction coeficients:
  if (TGoal(Goal).MoodCoef >= 0) then
    fStepReduc:= fStepReduc * Power(TGoal(Goal).MoodCoef, TGoal(Goal).Strength);
  // Acummulate reductions to apply at the end of the future (goblally reductive):
  if (TGoal(Goal).Gain <> 1) and
     (TGoal(Goal).ReactionMSeconds > 0) and
     (TimePassed < TGoal(Goal).ReactionMSeconds) then begin
    // GainCoef (from 0 to 1) will reduce the future score at the end, but we
    // need to make it less importat when it happend far in the future, so dying
    // in 1s is quite important, but dying in 6s is really not disturbing.
    //
    // To build this time coef. we use 2 params:
    //  - Reaction time = "start warning me if it happends in T < N seconds"
    //  - Then it can increase lineally with T or more dramatically.
    TimeCoef:= Power(1.0*TimePassed / TGoal(Goal).ReactionMSeconds, TGoal(Goal).Urgency);
    // The warning coef to be used needs to grow to 1 at reaction time:
    Gain:= TGoal(Goal).Gain;
    Warning:= Gain + (1-Gain) * TimeCoef;
    fStepGain:= fStepGain * Warning;
  end;
end;

// After a step is made the points earned are added to the future:
procedure TState.AddStepScoring;
var
  JustEnjoyed: double;
begin
  // As mentioned above on AddGoalScoring, StepEnjoy is half backed, it hold the
  // sum of the square of all goal's enjoys, to be like a euclidean distance
  // you still need to make the squared root of it (and multiply by Reduc):
  JustEnjoyed:= sign(fStepEnjoy) * Sqrt(abs(fStepEnjoy)) * fStepReduc;
  // Add them to the future:
  Enjoy:= Enjoy + JustEnjoyed * TPlayer(Owner).StepSeconds;
  // Gain feelings are also accumulated:
  GainCoef:= GainCoef * fStepGain;
  // Draw the point of the future just scored:
  TPlayer(Owner).DrawFuturePoint;
  // Reset scorings to prepare for the next time step:
  ResetStepScoring;
end;

function TState.GetDecisionVector: string;
var i: integer;
begin
  result:= '';
  for i:= 0 to FreeParamCount-1 do begin
    if (result<>'') then result:= result + ', ';
    result:= result + IntToStr(round(FreeParam[i].IniValue));
  end;
  result:= '(' +result+') #'+IntToStr(HitCount);
end;

// ********************
// **     TState     **
// ** THINKING STUFF **
// ********************

// Two futures with same initial decision belongs to the same "option" so they
// are two possible future endings for a single initial decision (option):
function TState.HaveSameOption(Other: TState): boolean;
begin
  result:= (OptionId = Other.OptionId);
end;

// Two futures that end "near" represent the same final state.
function TState.HaveSameFinalState(Other: TState): boolean;
var i: integer;
begin
  // Each Positional param (macro related param) must have identical value:
  for i:= 0 to PositionalParamCount-1 do begin
    if (PositionalParam[i].RoundValue <> Other.PositionalParam[i].RoundValue) then begin
      // They don't end in the same MacroState!
      result:= false;
      exit;
    end;
  end;
  // No difference was found, they share the final MacroState!
  result:= true;
end;

end.
