unit UTeam;

// TEAM is a group of players that are in a track playing.
// It can have a "Team Intelligence" that will guide the players into the team
// goals

interface

uses Contnrs, SysUtils, Math,
     UPlayer, UGoalSet, UGoal;

type
  TTeam = class(TObject)
  private
    // Track I am playing in:
    fTrack: TObject;
    // Player actually "selected":
    fSelectedPlayer: TPlayer;
    // The team members are imiagining futures, please call later
    fImaginig: boolean;
    // Witch things motivate this team (affect all its player):
    fGoalSet: TGoalSet;
    // Property getters
    procedure SetImagining(v: boolean);
    function  GetPlayer(i: integer): TPlayer;
    function  GetPlayerCount: integer;
    procedure SetSelectedPlayer(v: TPlayer);
    function  GetStepSeconds: double;
  protected
  public
    // List of al the players on track (as TObjects):
    fPlayers:   TObjectList;
    constructor Create;
    destructor  Destroy; override;
    procedure   AddPlayer(v: TPlayer);
    procedure   DelPlayer(v: TPlayer; DoFree: boolean);
    // Do things on all your team members (players):
    procedure   ResetFutures;
    procedure   ResetState;
    procedure   SetIniState;
    procedure   DrawOnFrame;
    procedure   TakeAStep(phase: integer);
    procedure   TakeFirstStep;
    procedure   TakeRandomStep;
    procedure   CalcGoals;
    procedure   TakeIntelligentStep;
    procedure   Duplicate;
    procedure   Draw;
    procedure   DrawOptions;
    property SelectedPlayer: TPlayer read fSelectedPlayer write SetSelectedPlayer;
    property Player[i: integer]: TPlayer read GetPlayer;
    property PlayerCount: integer read GetPlayerCount;
    property GoalSet: TGoalSet read fGoalSet write fGoalSet;
    property Track: TObject read fTrack write fTrack;
    property StepSeconds: double read GetStepSeconds;
    property Imagining: boolean read fImaginig write SetImagining;
  end;

implementation

uses UTrack, UTrack2D, UListOfFutures, UPlayer2D;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TTeam.Create;
begin
  inherited;
  fPlayers:= TObjectList.create;
  fPlayers.OwnsObjects:= false;
  fGoalSet:= TGoalSet.create(self);
  fSelectedPlayer:= nil;
  fImaginig:= false;
end;

destructor TTeam.destroy;
begin
  while fPlayers.Count > 0 do begin
    TPlayer(fPlayers.Items[0]).Free;
    fPlayers.Delete(0);
  end;
  fPlayers.Free;
  fGoalSet.Free;
  inherited;
end;


// ********************
// ** THINKING STUFF **
// ********************

// Make the intial random step for this new future:
procedure TTeam.TakeFirstStep;
begin
  TakeAStep(0);
end;

// Make another future's random step:
procedure TTeam.TakeRandomStep;
begin
  TakeAStep(1);
end;

// Make the real -not imagining a future- intelligent step:
procedure TTeam.TakeIntelligentStep;
begin
  TakeAStep(2);
end;

// Taking first step in a future is quite similar with taking the next steps
// or doing the final real step, so the 3 procedures use this general one:
//   phase = 0 : Initial future step.
//   phase = 1 : Following future's steps.
//   phase = 2 : Real step (not imagining futures).
procedure TTeam.TakeAStep(phase: integer);
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do begin
    // Store actual state as oldstate:
    Player[p].State.SetOldState;
    // Take a decision (first, random or intelligent decision):
    // Note: Take a decision start by storing old state values.
    case phase of
      0: Player[p].TakeFirstDecision;
      1: Player[p].TakeRandomDecision;
      2: Player[p].TakeIntelligentDecision;
    end;
    // Simulate a delta time so you update your internal state:
    Player[p].Simulate;
  end;
end;

function TTeam.GetStepSeconds: double;
begin
  result:= TTrack(Track).StepSeconds;
end;

procedure TTeam.CalcGoals;
var
  p: integer;
begin
  GoalSet.CalcGoals;
  for p:= 0 to PlayerCount-1 do
    Player[p].GoalSet.CalcGoals;
end;

procedure TTeam.DrawOnFrame;
var
  p: integer;
begin
  // Team goals:
  GoalSet.DrawOnFrame;
  // Players goals:
  for p:= 0 to PlayerCount-1 do
    Player[p].GoalSet.DrawOnFrame;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

// Ask players to duplicate if all is ok:
procedure TTeam.Duplicate;
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    if Player[p].CanDuplicate then
      Player[p].Clone;
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TTeam.Draw;
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].Draw;
end;

procedure TTeam.DrawOptions;
var p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].DrawOptions;
end;


// **********************
// ** MANAGING PLAYERS **
// **********************

// Accesing a member player:
function TTeam.GetPlayer(i: integer): TPlayer;
begin
  result:= TPlayer(FPlayers[i]);
end;

// Get number of players in the team:
function TTeam.GetPlayerCount: integer;
begin
  result:= FPlayers.Count;
end;

// Add a player to the team:
procedure TTeam.AddPlayer(v: TPlayer);
begin
  if fPlayers.IndexOf(v) = -1 then begin
    fPlayers.Add(v);
    v.Team:= Self;
    SelectedPlayer:= v;
    TTrack(Track).RefreshPlayerList;
  end;
end;

// Delete a player from the team:
procedure TTeam.DelPlayer(v: TPlayer; DoFree: boolean);
var p: integer;
begin
  p:= fPlayers.IndexOf(v);
  if (p <> -1) then begin
    // Avoid having a null selected player
    if (SelectedPlayer = v) then
      fSelectedPlayer:= nil;
    // Delete player:
    if DoFree then
      Player[p].free;
    fPlayers.Delete(p);
    // Chose another selected pplayer_
    if not Assigned(fSelectedPlayer) and (PlayerCount>0) then
      fSelectedPlayer:= Player[0];
    // Track be notified:
    TTrack(Track).RefreshPlayerList;
  end;
end;

// Witch player is selected by the user?
procedure TTeam.SetSelectedPlayer(v: TPlayer);
begin
  if fPlayers.IndexOf(v) = -1 then
    raise Exception.create('Selected player must be a team member.');
  fSelectedPlayer:= v;
end;


// *************************************
// ** ASK ALL YOUR PLAYERS TO DO THIS **
// *************************************

procedure TTeam.SetImagining(v: boolean);
var i: integer;
begin
  fImaginig:= v;
  for i:= 0 to PlayerCount-1 do
    Player[i].Imagining:= v;
  GoalSet.Reset;
end;

procedure TTeam.ResetFutures;
var i: integer;
begin
  for i:= 0 to PlayerCount-1 do
    Player[i].ResetFutures;
end;

procedure TTeam.SetIniState;
var
  p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].SetIniState;
end;

procedure TTeam.ResetState;
var
  p: integer;
begin
  for p:= 0 to PlayerCount-1 do
    Player[p].ResetState;
  GoalSet.Reset;
end;


end.
