unit UTrack;

// TRACK is the place where the simulation runs, the chessboard, it knows
// with teams and individual players exists, check for collisions and acts
// as a director on the thinking proccess by syncing all players' future
// creation.

interface

uses ExtCtrls, Contnrs, Forms, StdCtrls,
     UTeam;

type
  TTrack = class(TObject)
  private
    fTeams: TObjectList;
    fSelectedTeam: TTeam;
    // How many milliseconds in the future look ahedad when thinking?
    fFutureMSeconds: integer;
    fStepMSeconds: integer;
    // How many futures will I compute before deciding?
    fNumFutures: integer;
    // Detect collisions player-player?
    fPlayersCollide: boolean;
    // Activate evolution?
    fEvolution: boolean;
    // Property getters
    function  GetTeam(i: integer): TTeam;
    function  GetTeamCount: integer;
    function  GetPlayer(i: integer): TObject;
    function  GetPlayerCount: integer;
    function  GetStepSeconds: double;
    procedure SetEvolution(v: boolean);
  protected
    fPlayers: TObjectList;
    procedure DetectCollisions; virtual; abstract;
    procedure ClearTrack; virtual; abstract;
  public
    constructor Create;
    destructor  Destroy; override;
    // Make all the players to think and move for one tick:
    procedure TakeAStep;
    procedure DelTeam(v: TTeam; DoFree: boolean = false);
    // Draw track and all players on theirs positions:
    // Note: "Clear" image before drawing is an option here. During simulation,
    // it is cleared before starting calcs, during the calcs. debug info is added
    // and finally, when all is decided, players are added calling draw(false):
    procedure Draw(Phase: integer); // 0 = Previous to think, 1 = After think.
    procedure DrawOptions;
    procedure AddTeam(v: TTeam);
    procedure RefreshPlayerList;
    function  PlayerCollides(MyPlayer: TObject): boolean;
    property  Team[i: integer]: TTeam read GetTeam;
    property  TeamCount: integer read GetTeamCount;
    property  SelectedTeam: TTeam read fSelectedTeam write fSelectedTeam;
    property  Player[i: integer]: TObject read GetPlayer;
    property  PlayerCount: integer read GetPlayerCount;
    property  NumFutures: integer read fNumFutures write fNumFutures;
    property  FutureMSeconds: integer read fFutureMSeconds write fFutureMSeconds;
    property  StepMSeconds: integer read fStepMSeconds write fStepMSeconds;
    property  StepSeconds: double read GetStepSeconds;
    property  PlayersCollide: boolean read fPlayersCollide write fPlayersCollide;
    property  Evolution: boolean read fEvolution write SetEvolution;
  end;


implementation

uses DateUtils, SysUtils, Graphics, Types, Math,
     UUtils, UPlayer, UGoal;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TTrack.Create;
begin
  inherited;
  // Teams on track:
  fTeams:= TObjectList.create;
  fTeams.OwnsObjects:= false;
  fSelectedTeam:= nil;
  // Players on the track:
  // Note: It is redundant with team list, but checking collisions needs it.
  fPlayers:= TObjectList.create;
  fPlayers.OwnsObjects:= false;
  // Futures to compute and for how many seconds:
  fNumFutures:= 100;
  fFutureMSeconds:= 10*1000; // 10s default
  fStepMSeconds:= 100; // 0.1s default
  // Activate evolution?
  Evolution:= false;
end;

destructor TTrack.destroy;
begin
  FTeams.Free;
  FPlayers.free;
  inherited;
end;


// ********************
// ** MANAGING TEAMS **
// ********************

function TTrack.GetTeam(i: integer): TTeam;
begin
  result:= TTeam(fTeams[i]);
end;

function TTrack.GetTeamCount: integer;
begin
  result:= FTeams.Count;
end;

// Add a team to the track:
procedure TTrack.AddTeam(v: TTeam);
begin
  FTeams.Add(v);
  v.Track:= Self;
  SelectedTeam:= v;
end;

// Delete a team from the track:
procedure TTrack.DelTeam(v: TTeam; DoFree: boolean = false);
var p: integer;
begin
  p:= FTeams.IndexOf(v);
  if (p <> -1) then begin
    if DoFree then
      Team[p].free;
    FTeams.Delete(p);
  end;
end;


// ******************************
// ** LIST OF PLAYERS ON TRACK **
// ******************************

function TTrack.GetPlayerCount: integer;
begin
  result:= fPlayers.Count;
end;

function TTrack.GetPlayer(i: integer): TObject;
begin
  if (i<0) or (i>PlayerCount-1) then
    result:= nil
  else
    result:= TPlayer(fPlayers[i]);
end;

// Add/Del players using the track instead of the team, as the track needs the
// list of all players on it in order to detect collisions of any 2 players.
procedure TTrack.RefreshPlayerList;
var
  t, p: integer;
begin
  fPlayers.Clear;
  for t:= 0 to TeamCount-1 do begin
    for p:= 0 to Team[t].PlayerCount-1 do begin
      fPlayers.Add(Team[t].Player[p]);
    end;
  end;
end;

// To avoid placing one player over another, use this function:
function TTrack.PlayerCollides(MyPlayer: TObject): boolean;
var
  p: integer;
begin
  // Collides with track limits?
  result:= not TPlayer(MyPlayer).IsInsideTrack;
  if result then exit;
  // Collides with other players?
  for p:= 0 to PlayerCount-1 do begin
    if (Player[p] <> MyPlayer) then begin
      if TPlayer(MyPlayer).CheckCollision(TPlayer(Player[p])) then begin
        result:= true;
        exit;
      end;
    end;
  end;
end;

// ********************
// ** THINKING STUFF **
// ********************


// Make all players to think, decide, and move (do a complete frame):
// Note: All times in milliseconds.
procedure TTrack.TakeAStep;
var
  p, t, Ticks: integer;
  FoundFutures: integer;
  T0: TDateTime;
begin
  // Number of ticks needed to complete a future:
  Ticks:= Max(2, round(FutureMSeconds/StepMSeconds));
  // Reset the list of futures found:
  for p:= 0 to TeamCount-1 do
    Team[p].ResetFutures;
  // Clear the track image:
  Draw(0);
  // Compute all the needed futures:
  T0:= Now;
  FoundFutures:= 0;
  repeat
    // Start imagining a new future:
    for p:= 0 to TeamCount-1 do
      Team[p].Imagining:= true;
    // Take "sim_time" steps until you get to FutureSeconds
    for t:= 1 to ticks do begin
      // Ask each player to advance it's future a delta-time:
      for p:= 0 to TeamCount-1 do
        if (t=1) then
          // First step is different: it decides an "option", not a random one.
          Team[p].TakeFirstStep
        else
          Team[p].TakeRandomStep;
      // Detect and react to collisions:
      DetectCollisions;
      // Score changes on teams and players:
      for p:= 0 to TeamCount-1 do
        Team[p].CalcGoals;
      // Now sum this StepScore to the future's scoring (player by player):
      for p:= 0 to PlayerCount-1 do
        TPlayer(fPlayers[p]).State.AddStepScoring;
    end;
    // End imagining this future:
    for p:= 0 to TeamCount-1 do
      Team[p].Imagining:= false;
    // Another future computed!
    inc(FoundFutures);
  until (FoundFutures >= NumFutures) and (MilliSecondOfTheDay(Now-T0) > StepMSeconds);
  // Now players are back to reality and ready to make a decision and move on.
  //
  // Score state changes on all teams and all players in real life (not imagining)
  // so drops are actually collected, for instance:
  for p:= 0 to TeamCount-1 do begin
    Team[p].GoalSet.CalcGoals;
    // Draw goals that need drawing and don't draw on eachplayer:
    Team[p].DrawOnFrame;
  end;
  // End Drawing of the frame:
  Draw(1);
  // Analize the futures found and decide in a "intelligent" way:
  for p:= 0 to TeamCount-1 do
    Team[p].TakeIntelligentStep;
  // Detect "real" collisions between any two players and tell them to react:
  DetectCollisions;
  // Evolution?
  if Evolution then begin
    for p:= 0 to TeamCount-1 do
      Team[p].Duplicate;
  end;
end;

function TTrack.GetStepSeconds: double;
begin
  result:= StepMSeconds/1000;
end;

// ***************
// ** EVOLUTION **
// ***************

procedure TTrack.SetEvolution(v: boolean);
var
  p: integer;
begin
  fEvolution:= v;
  if Evolution then begin
    for p:= 0 to PlayerCount-1 do begin
      TPlayer(fPlayers[p]).GoalSet.ColorizeOwner;
      TPlayer(fPlayers[p]).GoalSet.Debug:= true;
    end;
  end;
end;

// *************
// ** DRAWING **
// *************

// Draw all players over the track:
procedure TTrack.Draw(Phase: integer);
var
  t: integer;
begin
  case Phase of
  -1: // Complete frame drawing:
    begin
      Draw(0);
      Draw(1);
    end;
  0: // Initial draw:
    begin
      //Repaint the backgroud:
      ClearTrack;
    end;
  1: // Final draw:
    begin
      // Draw players options (futures end points):
      for t:= 0 to TeamCount-1 do
        Team[t].DrawOptions;
      // Draw players on top of all:
      for t:= 0 to TeamCount-1 do
        Team[t].Draw;
      // Refresh screen:
      Application.ProcessMessages;
    end;
  end;
end;

procedure TTrack.DrawOptions;
var t: integer;
begin
  for t:= 0 to TeamCount-1 do
    Team[t].DrawOptions;
end;

end.
