unit UTrack2D;

interface

uses ExtCtrls, SysUtils,
     UTrack, UPlayer2D, UPhysic2D;

type
  TTrack2D = class(TTrack)
  private
    TrackScan: array of array of byte; //0=track, 1=out, 2=red recharge
    // General friction (avoid running too fast):
    fFriction: double;
    // Side friction (avoid sliding):
    fSideFriction: double;
    // Pointer to the track image so i can refresh it:
    fImage: TImage;
    // Image dimensions:
    fImageMaxX, fImageMaxY: integer;
    procedure ScanTrack(Im: TImage);
    function  GetPlayer2D(i: integer): TPhysic2D;
  protected
    procedure DetectCollisions; override;
  public
    CopyImage: TImage;
    constructor Create;
    function    MouseIn: boolean;
    function    MouseX: integer;
    function    MouseY: integer;
    function    Width: integer;
    function    Height: integer;
    function    IsInsideImage(x, y: double): boolean;
    function    IsInsideTrack(x, y: double): boolean;
    procedure   ClearTrack; override;
    property    Player2D[i: integer]: TPhysic2D read GetPlayer2D;
    property    Img: TImage read FImage write ScanTrack;
    property    Friction: double read FFriction write FFriction;
    property    SideFriction: double read FSideFriction write FSideFriction;
  end;

implementation

uses Graphics, Forms, Math, Controls,
     UTeam, UUTils, UKart, URocket;


// ***************************
// ** ADMNINISTRATIVE STUFF **
// ***************************

constructor TTrack2D.Create;
begin
  inherited;
  fImage:= Nil;
  fFriction:= 0.008;
  fSideFriction:= 0.10;
end;

function TTrack2D.GetPlayer2D(i: integer): TPhysic2D;
begin
  if fPlayers[i] is TPhysic2D then
    result:= TPhysic2D(fPlayers[i])
  else
    raise Exception.create('Track2D contains a player that is not a Physic2D player.');
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TTrack2D.ScanTrack(Im: TImage);
var
  i, j: integer;
begin
  // Remember it for later use:
  FImage:= Im;
  // Store dimensiones:
  FImageMaxX:= Im.Picture.Width;
  FImageMaxY:= Im.Picture.Height;
  // Scan picture for black pixles:
  SetLength(TrackScan, Im.Picture.Width);
  for i:= Low(TrackScan) to High(TrackScan) do begin
    SetLength(TrackScan[i], Im.Picture.Height);
    for j:= Low(TrackScan[i]) to High(TrackScan[i]) do begin
      TrackScan[i, j]:= 0;
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clBlack) then
        TrackScan[i, j]:= 1; // Outside track
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clRed) then
        TrackScan[i, j]:= 2; // Battery charging zone
    end;
  end;
  // I need a duplicate of the track so I won't paint over the original one.
  CopyImage:= TImage.Create(nil);
  CopyImage.Visible:= false;
  CopyImage.Width:=   Im.Width;
  CopyImage.Height:=  Im.Height;
  CopyImage.Left:=    Im.Left;
  CopyImage.Top:=     Im.Top;
  CopyImage.Picture.Assign(Im.Picture);
end;

procedure TTrack2D.ClearTrack;
begin
  if Assigned(FImage) then
    FImage.Picture.Assign(CopyImage.Picture);
end;

function TTrack2D.MouseIn: boolean;
var x, y: integer;
begin
  x:= MouseX;
  y:= MouseY;
  result:= (x>0) and (x<Img.Width) and (y>0) and (y<Img.Height);
end;

function TTrack2D.MouseX: integer;
begin
  result:= Img.ScreenToClient(Mouse.CursorPos).X;
end;

function TTrack2D.MouseY: integer;
begin
  result:= Img.ScreenToClient(Mouse.CursorPos).Y;
end;


function TTrack2D.Width: integer;
begin
  result:= Img.Width;
end;
function TTrack2D.Height: integer;
begin
  result:= Img.Height;
end;


// **********************
// ** SIMULATION STUFF **
// **********************

// Detect collisions: Player vs Player and Player vs Track and make them react:
procedure TTrack2D.DetectCollisions;
var
  p: array [1..2] of integer;
  pi, pj, k, m: integer;
  ImpX, ImpY: double;
  C2I_X, C2I_Y, C2I_L, I2Cn_X, I2Cn_Y: double;
  RF_X, RF_Y: double;
  Hit: boolean;
  Bounce, TotMass, Cross, Prod: double;
  IsKart, IsRocket: integer;
  MassRelation: double;
  signed: integer;
  VelX, VelY: double;
  AddVelX, AddVelY, AddVelA: array [1..2] of double;
  ActX, ActY, ActA: array [1..2] of extended;
  Momentum0, Usable: double;
const
  Debug: boolean = false;
  Elasticity: double = 1.75; // 1=Sticky, 2=Perfect elastic (avoid < 1.5!)

  function Player(ind: integer): TPhysic2D;
  begin
    result:= Player2D[p[ind]];
  end;
  function Other(ind: integer): TPhysic2D;
  begin
    result:= Player2D[p[3-ind]];
  end;

  // Kynetic energy of both colliding players if you use v * reactions:
  function Momentums(v: double): double;
  var
    kk: integer;
    vx, vy, va: extended;
  begin
    result:= 0;
    for kk:= 1 to 2 do begin
      vx:= Player(kk).VelX;
      vy:= Player(kk).VelY;
      va:= Player(kk).VelA;
      if (v>0) then begin
        vx:= vx + AddVelX[kk] * v * Elasticity;
        vy:= vy + AddVelY[kk] * v * Elasticity;
        va:= va + AddVelA[kk] * v * sqrt(Elasticity);
      end;
      result:= result +
        Player(kk).Mass * sqrt(sqr(vx/100) + sqr(vy/100)) +
        Player(kk).AngInertia * (abs(va)/100);
    end;
  end;

begin
  // Player vs Player collisions are possible (using bounding boxes)?
  if PlayersCollide then begin
    // Check all pairs:
    for pi:= 0 to PlayerCount-2 do begin
      for pj:= (pi+1) to PlayerCount-1 do begin
        // First check if a collision is even possible:
        if (Player2D[pi].MinX <= Player2D[pj].MaxX) and
           (Player2D[pi].MaxX >= Player2D[pj].MinX) and
           (Player2D[pi].MinY <= Player2D[pj].MaxY) and
           (Player2D[pi].MaxY >= Player2D[pj].MinY) then begin
          // Players are colliding? Gimme the contact point!
          if Player2D[pi].CheckCollision(Player2D[pj]) then begin
            // Store index of both colliding objects:
            p[1]:= pi;
            p[2]:= pj;
            // Separate them before bouncing:
            for k:= 1 to 2 do begin
              ActX[k]:= (Player2D[p[k]].State.ParamByName['PosX' ].MemoValue - Player2D[p[k]].PosX)  / 10;
              ActY[k]:= (Player2D[p[k]].State.ParamByName['PosY' ].MemoValue - Player2D[p[k]].PosY)  / 10;
              ActA[k]:= (Player2D[p[k]].State.ParamByName['Angle'].MemoValue - Player2D[p[k]].Angle) / 10;
            end;
            m:= 20; // 9 should be enough.
            repeat
              ImpX:= Player(1).ImpactX;
              ImpY:= Player(1).ImpactY;
              // --- debug ---
              if Debug and not Player2D[pi].Imagining then begin
                Player2D[pi].Draw;
                Player2D[pj].Draw;
                Img.Canvas.Brush.Style:= bsClear;
                Img.Canvas.Pen.Color:= clBlack;
                Img.Canvas.Ellipse(round(ImpX-10), round(ImpY-10), round(ImpX+10), round(ImpY+10));
                Application.ProcessMessages;
              end;
              for k:= 1 to 2 do begin
                Player(k).PosX:=  Player(k).PosX  + ActX[k];
                Player(k).PosY:=  Player(k).PosY  + ActY[k];
                Player(k).Angle:= Player(k).Angle + ActA[k];
              end;
              dec(m);
            until (m<0) or Player(1).CheckCollision(Player(2));
            // Initial kynetic energy >= Final kynetic energy (avoid quantic jumps!):
            Momentum0:= Momentums(0);
            // Make them react to the impact, one by one:
            for k:= 1 to 2 do begin
              // First thing I need is the vector from Player to impact point (p->i):
              C2I_X:= ImpX - Player(k).PosX;
              C2I_Y:= ImpY - Player(k).PosY;
              // --- debug ---
              if Debug and not Player(k).Imagining then begin
                Img.Canvas.Pen.Color:= clLime;
                Img.Canvas.MoveTo(round(Player(k).PosX), round(Player(k).PosY));
                Img.Canvas.LineTo(round(Player(k).PosX + C2I_X*5), round(Player(k).PosY + C2I_Y*5));
                Img.Canvas.Pen.Color:= clBlack;
                Img.Canvas.MoveTo(round(Player(k).PosX), round(Player(k).PosY));
                Img.Canvas.LineTo(round(Player(k).PosX + Player(k).VelX*5), round(Player(k).PosY + Player(k).VelY*5));
                Application.ProcessMessages;
              end;
              // Distance from players:
              C2I_L:= Sqrt(Sqr(C2I_X)+Sqr(C2I_Y));
              // No zero divisions:
              if (C2I_L = 0) then
                C2I_L:= 1;
              // Vector distance inversed and normalized:
              I2Cn_X:= -C2I_X / C2I_L;
              I2Cn_Y:= -C2I_Y / C2I_L;
              // Crashing Vel = Difference of velocities:
              VelX:= -(Player(k).VelX - Other(k).VelX);
              VelY:= -(Player(k).VelY - Other(k).VelY);
              // --- debug ---
              if Debug and not Player(k).Imagining then begin
                Img.Canvas.Pen.Color:= clGreen;
                Img.Canvas.MoveTo(round(Player(k).PosX), round(Player(k).PosY));
                Img.Canvas.LineTo(round(Player(k).PosX + VelX*5), round(Player(k).PosY + VelY*5));
                Application.ProcessMessages;
              end;
              // Part of the reaction force applied to the velocity:
              Cross:= (VelX * I2Cn_X + VelY * I2Cn_Y);
              Bounce:=  Cross * Other(k).Mass / (Other(k).Mass + Player(k).Mass);
              AddVelX[k]:= (Bounce * I2Cn_x);
              AddVelY[k]:= (Bounce * I2Cn_y);
              // --- debug ---
              if Debug and not Player(k).Imagining then begin
                Img.Canvas.Pen.Color:= clBlue;
                Img.Canvas.MoveTo(round(Player(k).PosX), round(Player(k).PosY));
                Img.Canvas.LineTo(round(Player(k).PosX + AddVelX[k]*5), round(Player(k).PosY + AddVelY[k]*5));
                Application.ProcessMessages;
              end;
              // Rest of reaction Vel? (vector normal to C2I)
              VelX:= VelX - AddVelX[k];
              VelY:= VelY - AddVelY[k];
              // The rest of reaction force is converted into torque:
              // Length of N (normal velocity) / radio = angular velocity to add:
              // Note: N is the z-coordinte on vectorial product (C2In ^ Vel).
              Prod:= (I2Cn_x * VelY) - (I2Cn_y * VelX);
              AddVelA[k]:= Prod / Player(k).AngInertia;
              // Bounce speed squared = bouncing energy, take it from both players' energy:
              // Note: Mass not used, integrity forces supposed to be proportional (but they arent)
              Player(k).Health:= Player(k).Health - sqr(Bounce);
            end;
            //
            // Bouncing on both players is calculated, adjust it to avoid problemas:
            //
            // Avoid creating/destroying energy by reducing boucing velocities:
            Usable:= 1;
            while (Momentums(Usable) > Momentum0) and (Usable>0.5) do
              Usable:= Usable * 0.9;
            // Reduce speed if neccesary to avoid energy creation:
            if (Usable < 1) then begin
              for k:= 1 to 2 do begin
                AddVelX[k]:= AddVelX[k] * Usable;
                AddVelY[k]:= AddVelY[k] * Usable;
                AddVelA[k]:= AddVelA[k] * Usable;
              end;
            end;
            // Make player bounce taking off the inelasticity:
            for k:= 1 to 2 do begin
              Player(k).VelX:= Player(k).VelX + AddVelX[k] * Elasticity;
              Player(k).VelY:= Player(k).VelY + AddVelY[k] * Elasticity;
              Player(k).VelA:= Player(k).VelA + AddVelA[k] * sqrt(Elasticity);
            end;

            // Players can't touch after the frame:
            for k:= 1 to 2 do begin
              // If you have to walk back until not touching, use those deltas:
              ActX[k]:= (Player(k).PosX - ImpX) / 10;
              ActY[k]:= (Player(k).PosY - ImpY) / 10;
              ActA[k]:= 0; // AddVelA[k] / 10;
            end;
            // Avoid crashing on next frame (it would be a dissaster!):
            m:= 20; // Surely 8-9 is ok, but they MUST not collide at any cost.
            repeat
              for k:= 1 to 2 do begin
                // Walk back 1/10 of a time step:
                Player(k).PosX:=  Player(k).PosX  + ActX[k];
                Player(k).PosY:=  Player(k).PosY  + ActY[k];
                Player(k).Angle:= Player(k).Angle + ActA[k];
                // If it is leaving the track, stop it:
                if not Player(k).IsInsideTrack then begin
                  // Get out of the wall:
                  Player(k).PosX:=  Player(k).PosX  - ActX[k];
                  Player(k).PosY:=  Player(k).PosY  - ActY[k];
                  Player(k).Angle:= Player(k).Angle - ActA[k];
                  // Stop this one, give your moving to the other:
                  ActX[3-k]:= ActX[3-k] - ActX[k];
                  ActY[3-k]:= ActY[3-k] - ActY[k];
                  ActA[3-k]:= ActA[3-k] - ActA[k];
                  // ...so you stop moving:
                  ActX[k]:= 0;
                  ActY[k]:= 0;
                  ActA[k]:= 0;
                end;
                // ---Debug---:
                if Debug and not Player(k).Imagining then begin
                  Player(k).Draw;
                  Application.ProcessMessages;
                end;
              end;
              dec(m);
            until (m<=0) or not Player(1).CheckCollision(Player(2));
            //
            // Evolutive stuff: Karts eat rockets!
            if Evolution then begin
              // Kart touch rocket: Kart takes some energy from rocket:
              if (not Player(1).Dead) and (not Player(2).Dead) then begin
                IsKart:=  -1;
                IsRocket:= -1;
                if (Player(1) is TKart)   then IsKart:= 1;
                if (Player(2) is TKart)   then IsKart:= 2;
                if (Player(1) is TRocket) then IsRocket:= 1;
                if (Player(2) is TRocket) then IsRocket:= 2;
                if (IsKart<>-1) and (IsRocket<>-1) then begin
                  // Karts doesn't damage when crashing on a rockeck:
                  Player(IsKart).Health:= Player(IsKart).Health + Bounce;
                  // Suck some energy from the rocket ONLY if it is alive, or if
                  // I am imagining a future (I don't know if it is dead):
                  if (not Player(IsRocket).Dead) or (Player(IsRocket).Imagining) then begin
                    // Energy to transfer from rocket to kart: 25%
                    Bounce:= Min(25, Player(IsRocket).Energy);
                    Player(IsKart).Energy:=   Player(IsKart).Energy   + Bounce * 0.9; // 90% transfer eficiencie
                    Player(IsRocket).Energy:= Player(IsRocket).Energy - Bounce;
                  end;
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  end;
  // Check collisions player-track limits:
  for pi:= 0 to PlayerCount-1 do begin
    Player2D[pi].IsInsideTrack;
    Player2D[pi].BounceWithTrack;
  end;
end;

function TTrack2D.IsInsideImage(x, y: double): boolean;
var i, j: integer;
begin
  i:= round(x);
  j:= round(y);
  result:= (i>0) and (i<FImageMaxX) and (j>0) and (j<FImageMaxY);
end;

function TTrack2D.IsInsideTrack(x, y: double): boolean;
var i, j: integer;
begin
  i:= round(x);
  j:= round(y);
  if IsInsideImage(i, j) then
    result:= (TrackScan[i,j]<>1)
  else
    result:= false;
end;


end.
